"""
GillesPy2 is a modeling toolkit for biochemical simulation.
Copyright (C) 2019-2021 GillesPy2 developers.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from gillespy2.core import Model, Species, Reaction, Parameter
import numpy as np


class Example(Model):
    """
    This is a simple example for mass-action degradation of species S.
    """

    def __init__(self, parameter_values=None):
        # Initialize the model.
        Model.__init__(self, name="Example")
        # Species
        S = Species(name='Sp', initial_value=100)
        self.add_species([S])
        # Parameters
        k1 = Parameter(name='k1', expression=3.0)
        self.add_parameter([k1])
        # Reactions
        rxn1 = Reaction(name='S degradation', reactants={S: 1}, products={}, rate=k1)
        self.add_reaction([rxn1])
        self.timespan(np.linspace(0, 20, 101))


__all__ = ['Trichloroethylene', 'LacOperon', 'Schlogl', 'MichaelisMenten',
           'ToggleSwitch', 'Example', 'Tyson2StateOscillator', 'Oregonator']



"""
GillesPy2 is a modeling toolkit for biochemical simulation.
Copyright (C) 2019-2021 GillesPy2 developers.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from gillespy2.core.sortableobject import SortableObject
from gillespy2.core.gillespyError import *
from gillespy2.core.jsonify import Jsonify
import numpy as np

class Species(SortableObject, Jsonify):
    """
    Chemical species. Can be added to Model object to interact with other
    species or time.

    :param name: The name by which this species will be called in reactions and within the model.
    :type name: str

    :param initial_value: Initial population of this species. If this is not provided as an int,
        the type will be changed when it is added by numpy.int
    :type initial_value: int >= 0

    :param constant: If true, the value of the species cannot be changed (currently TauHybridSolver only)
    :type constant: bool

    :param boundary_condition: If true, species can be changed by events and rate rules, but not by reactions.
        (TauHybridSolver only)
    :type boundary_condition: bool

    :param mode: ***FOR USE WITH BasicTauHybridSolver ONLY***
        Sets the mode of representation of this species for the TauHybridSolver,
        can be discrete, continuous, or dynamic.
        mode='dynamic' - Allows a species to be represented as either discrete or continuous
        mode='continuous' - Species will only be represented as continuous
        mode='discrete' - Species will only be represented as discrete
    :type mode: str

    :param allow_negative_populations: If true, population can be reduces below 0.
    :type allow_negative_populations: bool

    :param switch_tol: ***FOR USE WITH BasicTauHybridSolver ONLY***
        Tolerance level for considering a dynamic species deterministically, value is compared to an estimated sd/mean
        population of a species after a given time step. This value will be used if a switch_min is not provided.
        The default value is 0.03
    :type switch_tol: float

    :param switch_min:  ***FOR USE WITH BasicTauHybridSolver ONLY***
        Minimum population value at which species will be represented as continuous. If a value is given, switch_min will be
        used instead of switch_tol
    :type switch_min: float
    """

    def __init__(self, name="", initial_value=0, constant=False,
                 boundary_condition=False, mode=None,
                 allow_negative_populations=False, switch_min=0,
                 switch_tol=0.03):
        # A species has a name (string) and an initial value (positive integer)
        self.name = name
        self.constant = constant
        self.boundary_condition = boundary_condition
        self.mode = mode
        self.allow_negative_populations = allow_negative_populations
        self.switch_min = switch_min
        self.switch_tol = switch_tol

        mode_list = ['continuous', 'dynamic', 'discrete', None]

        if self.mode not in mode_list:
            raise SpeciesError('Species mode must be either \'continuous\', \'dynamic\', \'discrete\', or '
                               '\'unspecified(default to dynamic for BasicTauHybridSolver)\'.')
        if mode == 'continuous':
            self.initial_value = np.float(initial_value)
        else:
            if np.int(initial_value) != initial_value:
                raise ValueError(
                    "'initial_value' for Species with mode='discrete' must be an integer value. Change to "
                    "mode='continuous' to use floating point values.")
            self.initial_value = np.int(initial_value)
        if not allow_negative_populations:
            if self.initial_value < 0:
                raise ValueError('A species initial value must be non-negative unless allow_negative_populations=True')

    def __str__(self):
        print_string = self.name
        print_string += ': ' + str(self.initial_value)
        '''
        print_string += '\n\tInitial Value: ' + str(self.initial_value)
        print_string += '\n\tConstant: ' + str(self.constant)
        print_string += '\n\tBoundary Condition: ' + str(self.boundary_condition)
        print_string += '\n\tMode: ' + self.mode
        print_string += '\n\tAllow Negative Populations: ' + str(self.allow_negative_populations)
        '''
        return print_string

    def set_initial_value(self, num):
        """
        Setter method for initial_value of a population

        :param num: Integer to set initial species population
        :raises SpeciesError: If num is non-negative or a decimal number
        """
        if isinstance(num, float) and (self.mode != 'dynamic' or self.mode != 'continuous'):
            raise SpeciesError("Mode set to discrete, species must be an integer number.")
        if num < 0 and self.allow_negative_populations == False:
            raise SpeciesError("Species population must be non-negative, or allow_negative_populations "
                               "must be set to True")
        self.initial_value = num


"""
GillesPy2 is a modeling toolkit for biochemical simulation.
Copyright (C) 2019-2021 GillesPy2 developers.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import sys
if sys.version_info < (3, 0):
    raise Exception("GillesPy2 only works in Python 3.0 and higher")
from .__version__ import __version__, __title__, __description__, __url__
from .__version__ import __author__, __email__
from .__version__ import __license__, __copyright__

from gillespy2.solvers import *
from gillespy2.core import *
import gillespy2.sbml

cache_enabled = False


"""
GillesPy2 is a modeling toolkit for biochemical simulation.
Copyright (C) 2019-2021 GillesPy2 developers.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from pathlib import Path
import unittest
import shutil
import os
import example_models

import gillespy2
from gillespy2.solvers.cpp.build.build_engine import BuildEngine


class TestBuildEngine(unittest.TestCase):
    test_model = example_models.Dimerization()

    # List of solver names.
    # Each entry represents the Makefile target for a C++ solver.
    solver_names = [
        "ssa",
        "ode",
        "tau_leap",
    ]

    def setUp(self) -> None:
        # Disable the gillespy2 cache.
        gillespy2.cache_enabled = False

        self.build_engine = BuildEngine()
        self.tmp_dir = self.build_engine.prepare(self.test_model, variable=False)

    def tearDown(self) -> None:
        self.build_engine.clean()

        if self.tmp_dir is not None and os.path.exists(self.tmp_dir):
            shutil.rmtree(self.tmp_dir, ignore_errors=True)

    def test_default_layout(self):
        """
        Ensure that with the default, cache-less output, the template and obj files
          are subdirectories of the temp directory.
        This is to ensure that they get cleaned up when the temp directory is cleaned up.
        """
        # This test should be attempted with each solver target.
        for solver_name in self.solver_names:
            with self.subTest(solver=solver_name):
                self.build_engine.build_simulation(solver_name)

        template_dir = self.build_engine.template_dir
        obj_dir = self.build_engine.obj_dir

        self.assertTrue(str(template_dir.resolve()).startswith(str(Path(self.tmp_dir).resolve())),
                        "Template directory was not placed in temp directory when cache is disabled")
        self.assertTrue(str(obj_dir.resolve()).startswith(str(Path(self.tmp_dir).resolve())),
                        "Dependency obj directory was not placed in temp directory when cache is disabled")

        # Ensure that the dependencies are cleaned up when .clean() is called.
        self.build_engine.clean()
        self.assertFalse(template_dir.exists(),
                         "Template directory was not cleaned up when cache is disabled")
        self.assertFalse(obj_dir.exists(),
                         "Dependency obj directory was not cleaned up when cache is disabled")

    def test_clean(self):
        """
        Ensure that with the default, cache-less output, the output files are cleaned up.
        """
        # Test clean-up with manual call to .clean()
        self.build_engine.clean()
        self.assertFalse(os.path.exists(self.tmp_dir), "Build engine output directory not cleaned up.")

    def test_template_file(self):
        """
        Ensure that the resulting template file exists and was prepared correctly.
        """
        template_file = self.build_engine.template_dir.joinpath(
                        self.build_engine.template_definitions_name)
        self.assertTrue(template_file.exists(), "Simulation template header file could not be found")

        # Test to ensure the contents of the output template_file contains ONLY macro definitions,
        # nothing else.
        with open(template_file) as template:
            for line in template.readlines():
                self.assertTrue(line.startswith("#"), "Output template file contains an invalid definition.")

    def test_build_output(self):
        """
        Ensure that the default, cache-less output has expected behavior.
        Should result in an expected definitions file and an executable simulation.
        """
        simulation_file = self.build_engine.output_dir.joinpath(
            "Simulation.exe" if os.name == "nt" else "Simulation.out")

        for solver_name in self.solver_names:
            with self.subTest(solver=solver_name):
                self.build_engine.build_simulation(solver_name)

        self.assertTrue(simulation_file.exists(), "Simulation executable could not be found")
        self.assertTrue(os.access(str(simulation_file), os.X_OK), "Simulation output is not executable")


if __name__ == '__main__':
    unittest.main()



import unittest

import gillespy2
import numpy
from gillespy2 import TauHybridCSolver
from example_models import Example, ExampleNoTspan
from gillespy2.core.gillespyError import SimulationError


class BoundaryConditionTestModel(gillespy2.Model):
    def __init__(self):
        gillespy2.Model.__init__(self, name="BoundaryConditionTestModel")
        s1 = gillespy2.Species(name="S1", boundary_condition=True, initial_value=0.001, mode='continuous')
        s2 = gillespy2.Species(name="S2", boundary_condition=False, initial_value=0.002, mode='continuous')
        s3 = gillespy2.Species(name="S3", boundary_condition=False, initial_value=0.001, mode='continuous')
        s4 = gillespy2.Species(name="S4", boundary_condition=True, initial_value=0.002, mode='continuous')
        self.add_species([s1, s2, s3, s4])

        k1 = gillespy2.Parameter(name="k1", expression="0.75")
        k2 = gillespy2.Parameter(name="k2", expression="0.25")
        self.add_parameter([k1, k2])

        reaction1 = gillespy2.Reaction(name="reaction1", propensity_function="vol*k1*S1*S2",
                                       reactants={s1: 1, s2: 1},
                                       products={s3: 1})
        reaction2 = gillespy2.Reaction(name="reaction2", propensity_function="vol*k2*S3",
                                       reactants={s3: 1},
                                       products={s1: 1, s2: 1})
        reaction3 = gillespy2.Reaction(name="reaction3", propensity_function="vol*k2*S3",
                                       reactants={s1: 1, s2: 2, s3: 1},
                                       products={s4: 1})
        self.add_reaction([reaction1, reaction2, reaction3])

        self.timespan(numpy.linspace(0, 20, 51))


# The output has a roughly predictable output, each species (approx.) approaching a particular value:
#   S1 -> 0     S2 -> 0.15      S3 -> 1.0 (should be approx. constant throughout)
class RateRuleTestModel(gillespy2.Model):
    def __init__(self):
        gillespy2.Model.__init__(self, name="RateRuleTestModel")
        s1 = gillespy2.Species(name="S1", initial_value=0.015, mode='continuous')
        s2 = gillespy2.Species(name="S2", initial_value=0.0, mode='continuous')
        s3 = gillespy2.Species(name="S3", initial_value=1.0, mode='continuous')
        self.add_species([s1, s2, s3])

        k1 = gillespy2.Parameter(name="k1", expression="1.0")
        self.add_parameter([k1])

        rule1 = gillespy2.RateRule(name="rule1", variable="S1", formula="-k1*S1")
        rule2 = gillespy2.RateRule(name="rule2", variable="S2", formula="k1*S1")
        rule3 = gillespy2.RateRule(name="rule3", variable="S3", formula="0.015-(S1+S2)")
        self.add_rate_rule([rule1, rule2, rule3])

        self.timespan(numpy.linspace(0, 20, 51))


class TestTauHybridCSolver(unittest.TestCase):
    def test_rate_rules(self):
        """
        Ensure that models provided with SBML rate rules work properly.
        Rate rule expressions are assumed valid and are tested in the Expressions test case.
        """
        model = RateRuleTestModel()
        results = model.run(solver=TauHybridCSolver, number_of_trajectories=1)

        self.assertAlmostEqual(results["S1"][-1], 0.0, places=3)
        self.assertAlmostEqual(results["S2"][-1], 0.015, places=3)
        self.assertTrue(numpy.all(results["S3"] == 1.0))

    def test_boundary_conditions(self):
        """
        Ensure that species marked with `boundary_condition=True` work properly.
        Boundary condition species should have a constant value throughout the simulation.
        Non-boundary condition species should change with the expected reaction rate.
        """
        model = BoundaryConditionTestModel()
        results = model.run(solver=TauHybridCSolver, number_of_trajectories=1)

        for spec_name, species in model.listOfSpecies.items():
            with self.subTest(msg="Unexpected species output for boundary condition setting",
                              boundary_condition=species.boundary_condition):
                spec_results = results[spec_name]
                is_uniform = numpy.all(spec_results == species.initial_value)
                self.assertTrue(species.boundary_condition == is_uniform)

    def test_run_example__with_increment_only(self):
        model = ExampleNoTspan()
        results = TauHybridCSolver.run(model=model, increment=0.2)

    def test_run_example__with_tspan_only(self):
        model = Example()
        results = TauHybridCSolver.run(model=model)

    def test_run_example__with_tspan_and_increment(self):
        with self.assertRaises(SimulationError):
            model = Example()
            results = TauHybridCSolver.run(model=model, increment=0.2)


if __name__ == '__main__':
    unittest.main()



"""
GillesPy2 is a modeling toolkit for biochemical simulation.
Copyright (C) 2019-2021 GillesPy2 developers.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import os
import sys

sys.path.append(os.path.join(os.path.dirname(__file__), '..'))

import unittest
from unittest import TestCase

from profiling import python_profiler

from example_models import Oregonator
from example_models import VilarOscillator
from example_models import Tyson2StateOscillator

from gillespy2.solvers.numpy.ode_solver import ODESolver
from gillespy2.solvers.numpy.ssa_solver import NumPySSASolver
from gillespy2.solvers.numpy.tau_leaping_solver import TauLeapingSolver


class TestPythonSolverPerf(TestCase):
    def setUp(self) -> None:
        self.solvers = {
            NumPySSASolver: [
                Tyson2StateOscillator()
            ],
            ODESolver: [
                Oregonator()
            ],
            TauLeapingSolver: [
                VilarOscillator()
            ],
        }

    def test_python_solver_perf(self):
        for solver, models in self.solvers.items():
            print(f"=== === === {solver.name} === === ===")

            for model in models:
                print(f"{model.name}:")

                perf_data = python_profiler.run_profiler(model, solver)
                print(perf_data)

            print()

        print()

if __name__ == "__main__":
    unittest.main()



"""
GillesPy2 is a modeling toolkit for biochemical simulation.
Copyright (C) 2019-2021 GillesPy2 developers.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import io
import abc
import numpy
from collections import deque
from typing import Callable
from abc import ABC

class SimDecoder(ABC):
    """
    Abstract simulation decoder class.
    For solvers to handle output in a custom way, this decoder class will be implemented.
    Expects the output to be read from a buffered IO reader.

    :param trajectories: 3D array to output simulation data to.
    :type trajectories: numpy.array
    """
    
    def __init__(self, trajectories: numpy.ndarray):
        """
        Constructor for simulation decoder class.
        Reads the output of an external simulation and generates a numpy array from the results.
        """
        # Make sure that it's actually a numpy array
        if not isinstance(trajectories, numpy.ndarray):
            raise io.UnsupportedOperation("Could not construct decoder: only NumPy arrays are supported")
        # ...and also that it's actually 3-dimensional
        if not len(trajectories.shape) == 3:
            raise io.UnsupportedOperation("Could not construct decoder: supplied NumPy array must be 3-dimensional")

        # Provided trajectory assumed valid
        self.trajectories = trajectories
        self.num_trajectories, self.num_timesteps, self.num_species = trajectories.shape
        self.bytes_read = 0

    @classmethod
    def create_default(cls, num_trajectories: int, num_timesteps: int, num_species: int) -> "SimDecoder":
        """
        Creates a new instance of the calling class, using a NumPy array with a predefined shape.
        Calling this method is preferred over calling the constructor directly.

        :param num_trajectories: Number of trajectories expected in the simulation output.
        :type num_trajectories: int

        :param num_timesteps: Number of timesteps expected in the simulation output.
        :type num_timesteps: int

        :param num_species: Number of species expected in the simulation output.
        :type num_species: int

        :returns: An instance of the decoder object, automatically populated with a valid output array.
        """
        return cls(numpy.zeros((num_trajectories, num_timesteps, num_species + 1)))

    @abc.abstractmethod
    def read(self, output: io.BufferedReader, **kwargs):
        """
        Accepts a buffered reader from stdout of a subprocess.
        Contents of the given reader are processed and made available through get_output().

        Blocks until the output of the buffered reader has been read completely.

        :param output: Reader provided from the stdout member of an open Popen class.
        :type output: io.BufferedReader
        """
        pass

    @abc.abstractmethod
    def get_output(self) -> "tuple[numpy.ndarray, int]":
        """
        Returns the fully-populated NumPy array containing the completed simulation data.
        Assumes that the subprocess has already completed.

        :returns: Tuple containing the 3D NumPy array of results, and the time stopped.
        """
        pass

class BasicSimDecoder(SimDecoder):
    """
    Simple decoder which returns the results as a complete string.
    """
    def __init__(self, trajectories: numpy.ndarray):
        super(BasicSimDecoder, self).__init__(trajectories)
        self.buffer = []

    def __read_next(self, output: io.BufferedReader):
        """
        Reads the next block from the simulation output.
        Returns the length of the string read.

        :param output: Reader provided from the stdout member of an open Popen object.
        :type output: io.BufferedReader

        :returns: Integer representing the number of characters read.
        """
        line = output.read().decode("utf-8")
        ln = len(line)
        if ln > 0:
            self.buffer.append(line)
        return ln

    def read(self, output: io.BufferedReader, **kwargs):
        page_size = self.__read_next(output)
        self.bytes_read = page_size
        while page_size > 0 and not output.closed:
            page_size = self.__read_next(output)
            self.bytes_read += page_size
        return self.bytes_read

    def get_output(self):
        stdout = "".join(self.buffer).split(",")
        # The last number written to stdout from C++ sim is always the stop time.
        time_stopped = stdout.pop()
        time_stopped = int(time_stopped) if time_stopped.isdigit() else 0

        # Assumed layout of NumPy array:
        #  1D: index to each simulation trajectory
        #  2D: index to each timestep of that directory
        #  3D: index to each species of that timestep
        # Buffer is a flat 1D list, which gets mapped into the NumPy array.
        for entry_i, entry in enumerate(stdout):
            spec_num = entry_i % self.num_species

            # Each timestep has a "stride" equal to the total number of species.
            entry_i //= self.num_species
            ts_num = entry_i % self.num_timesteps

            # Each trajectory has a "stride" equal to the total number of timesteps.
            traj_num = entry_i // self.num_timesteps

            self.trajectories[traj_num][ts_num][spec_num] = entry

        return self.trajectories, time_stopped


class IterativeSimDecoder(SimDecoder):
    """
    Output decoder for processing the output at regular intervals.
    An IterativeSimDecoder object can be provided a callback.
    Each time a new timestep is received and processed, the callback is invoked,
      and is provided with the timestep value and output time.

    Intended for handling just-in-time output.
    """

    def __init__(self, trajectories: numpy.ndarray, callback=None):
        super(IterativeSimDecoder, self).__init__(trajectories)
        self.callback = callback if callback is not None else lambda *args: None
        self.end_time = 0

    def with_callback(self, callback: "Callable[[int, numpy.ndarray], None]") -> "IterativeSimDecoder":
        """
        Provide a callback function to be invoked on each timestep.
        Accepts a function with the signature:

        def callback(float, numpy.ndarray)

        The first value (float) is the time value for the given timestep entry.
        The second value (NumPy array) is the simulation state for that timestep.
        Return values are ignored.

        :param callback: Function to be invoked on each timestep.
        :type callback: Callable[[int, numpy.ndarray], None]

        :returns: Pass-through for IterativeSimDecoder.
        """
        if not callable(callback):
            raise ValueError(f"Expected function as callback: got {type(callback)}")
        self.callback = callback
        return self

    def read(self, output: io.BufferedReader, page_size=256, **kwargs):
        """
        Read and process output from the provided buffer, one timestep at a time.
        Any registered callbacks will be invoked on each iteration of output processing.

        Blocks until the output of the buffered reader has been read completely.

        :param output: Reader provided from the stdout member of an open Popen class.
        :type output: io.BufferedReader

        :param page_size: Suggested value for number of bytes to read from the simulation on each pass.
        Smaller values may result in more consistent callback times, at the cost of performance.
        Larger values may result in better overall performance, at the cost of sporadic callback times.
        :type page_size: int

        :returns: Total number of bytes read
        """
        if page_size < 1:
            page_size = 256
        traj_id, t = 0, 0
        current_timestep = deque()
        num_trajectories, num_timesteps, entries_per_timestep = self.trajectories.shape

        line = output.read(page_size).decode("ascii")
        carry_value = ""
        bytes_read = len(line)
        self.bytes_read = bytes_read
        while bytes_read > 0:
            # carry_value is carried over from the previous read.
            # It must be appended to the beginning of the next read.
            if carry_value != "":
                line = carry_value + line
            entries = line.split(",")

            # The last "value" is not guaranteed to be a complete number.
            # It may be the start of an incomplete value. Example:
            # [ current read | next read ]
            # [...12,13,14,1 | 5,16,17...]
            # The value at the "boundary" should be 15, but if we do not carry,
            #   then it will be misread as two separate values, 1 and 5.
            carry_value = entries.pop()

            # Process incoming text, one timestep entry at a time.
            current_timestep.extend(entries)
            while len(current_timestep) >= entries_per_timestep:
                self.trajectories[traj_id][t] = [current_timestep.popleft() for _ in range(entries_per_timestep)]
                # First value of each timestep is the current time of the simulation.
                # The remaining entries of each timestep are the output state at that time.
                self.callback(self.trajectories[traj_id][t][0],
                              self.trajectories[traj_id][t][1:])
                t += 1
                if t >= num_timesteps:
                    traj_id += 1
                    t = 0

            # Process the next output block.
            # An empty output block indicates that the simulation has ended.
            line = output.read(page_size).decode("ascii")
            bytes_read = len(line)
            self.bytes_read += bytes_read

        if carry_value != "":
            self.end_time = int(carry_value)
        elif len(current_timestep) > 0:
            self.end_time = int(current_timestep.popleft())
        else:
            self.end_time = 0

        return self.bytes_read

    def get_output(self) -> "tuple[numpy.ndarray, int]":
        # TODO: block get_output() call if waiting on read() to complete
        return self.trajectories, self.end_time



"""
GillesPy2 is a modeling toolkit for biochemical simulation.
Copyright (C) 2019-2021 GillesPy2 developers.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

"""Class and methods for the Tau Leaping Solver"""
import random
import math
from threading import Thread, Event
import numpy as np
from gillespy2.solvers.utilities import Tau
from gillespy2.solvers.utilities import solverutils as nputils
from gillespy2.core import GillesPySolver, log, liveGraphing
from gillespy2.core import ModelError, ExecutionError
from gillespy2.core.results import Results

class TauLeapingSolver(GillesPySolver):
    """
    A Tau Leaping Solver for GillesPy2 models.  This solver uses an algorithm calculates
    multiple reactions in a single step over a given tau step size.  The change in propensities
    over this step are bounded by bounding the relative change in state, yielding greatly improved
    run-time performance with very little trade-off in accuracy.
    """

    name = "TauLeapingSolver"
    rc = 0
    stop_event = None
    pause_event = None
    result = None

    def __init__(self, model=None, debug=False, profile=False):
        name = "TauLeapingSolver"
        rc = 0
        stop_event = None
        pause_event = None
        result = None
        self.model = model
        self.debug = debug
        self.profile = profile

    def __get_reactions(self, step, curr_state, curr_time, save_time, propensities, reactions):
        """
        Helper Function to get reactions fired from t to t+tau.  
        :returns: Three values:
            rxn_count - dict with key=Reaction channel value=number of times fired
            curr_state - dict containing all state variables for system at current time
            curr_time - float representing current time
        """

        if curr_time + step > save_time:
            if self.debug:
                print("Step exceeds save_time, changing step size from ", step,
                      " to ", save_time - curr_time)
            step = save_time - curr_time

        if self.debug:
            print("Curr Time: ", curr_time, " Save time: ", save_time, "step: ", step)

        rxn_count = {}

        for rxn in reactions:
            rxn_count[rxn] = np.random.poisson(propensities[rxn] * step)

        if self.debug:
            print("Reactions Fired: ", rxn_count)

        curr_time = curr_time+step

        return rxn_count, curr_state, curr_time

    @classmethod
    def get_solver_settings(self):
        """
        :returns: Tuple of strings, denoting all keyword argument for this solvers run() method.
        """
        return ('model', 't', 'number_of_trajectories', 'increment', 'seed', 'debug', 'profile','timeout', 'tau_tol')

    @classmethod
    def run(self, model=None, t=20, number_of_trajectories=1, increment=None, seed=None,
            debug=False, profile=False,  live_output=None, live_output_options={},
            timeout=None, resume=None, tau_tol=0.03, **kwargs):
            """
            Function calling simulation of the model.
            This is typically called by the run function in GillesPy2 model objects
            and will inherit those parameters which are passed with the model
            as the arguments this run function.

            :param model: GillesPy2 model object to simulate
            :type model: gillespy2.Model

            :param t: Simulation run time
            :type t: int

            :param number_of_trajectories: Number of trajectories to simulate
            :type number_of_trajectories: int

            :param increment: Save point increment for recording data
            :type increment: float

            :param seed: The random seed for the simulation. Optional, defaults to None
            :type seed: int

            :param debug: Set to True to provide additional debug information about the simulation
            :type debug: bool

            :param profile: Set to True to provide information about step size (tau) taken at each step.
            :type profile: bool

            :param live_output: The type of output to be displayed by solver. Can be "progress", "text", or "graph".
            :type live_output: str

            :param live_output_options: COntains options for live_output. By default {"interval":1}. "interval"
                specifies seconds between displaying. "clear_output" specifies if display should be refreshed with each
                display.
            :type live_output_options: dict

            :param timeout:
            :param resume:
            :param tau_tol:
            :param kwargs:

            :returns:
            """

            if isinstance(self, type):
                self = TauLeapingSolver(model=model, debug=debug, profile=profile)

            increment = self.get_increment(model=model, increment=increment)

            self.stop_event = Event()
            self.pause_event = Event()

            if timeout is not None and timeout <= 0:
                timeout = None
            if len(kwargs) > 0:
                for key in kwargs:
                    log.warning('Unsupported keyword argument to {0} solver: {1}'.format(self.name, key))

            # create numpy array for timeline
            if resume is not None:
                # start where we last left off if resuming a simulatio
                lastT = resume['time'][-1]
                step = lastT - resume['time'][-2]
                timeline = np.arange(lastT, t+step, step)
            else:
                timeline = np.linspace(0, t, int(round(t / increment + 1)))

            species = list(model._listOfSpecies.keys())
            trajectory_base, tmpSpecies = nputils.numpy_trajectory_base_initialization(model, number_of_trajectories,
                                                                                       timeline, species, resume=resume)

            # total_time and curr_state are list of len 1 so that __run receives reference
            if resume is not None:
                total_time = [resume['time'][-1]]
            else:
                total_time = [0]

            curr_state = [None]
            live_grapher = [None]

            sim_thread = Thread(target=self.___run, args=(model, curr_state, total_time, timeline, trajectory_base, tmpSpecies,
                                                          live_grapher,), kwargs={'t': t,
                                                                                  'number_of_trajectories':
                                                                                      number_of_trajectories,
                                                                                  'increment': increment, 'seed': seed,
                                                                                  'debug': debug, 'resume': resume,
                                                                                  'timeout': timeout, 'tau_tol': tau_tol
                                                                                  })
            try:
                time = 0
                sim_thread.start()
                if live_output is not None:
                    import gillespy2.core.liveGraphing
                    live_output_options['type'] = live_output
                    gillespy2.core.liveGraphing.valid_graph_params(
                        live_output_options)
                    if resume is not None:
                        resumeTest = True  # If resuming, relay this information to live_grapher
                    else:
                        resumeTest = False
                    live_grapher[
                        0] = gillespy2.core.liveGraphing.LiveDisplayer(model,
                                                                       timeline,
                                                                       number_of_trajectories,
                                                                       live_output_options,
                                                                       resume=resumeTest)
                    display_timer = gillespy2.core.liveGraphing.RepeatTimer(
                        live_output_options['interval'],
                        live_grapher[0].display, args=(curr_state,
                                                       total_time,
                                                       trajectory_base,
                                                       live_output
                                                       )
                        )
                    display_timer.start()

                if timeout is not None:
                    while sim_thread.is_alive():
                        sim_thread.join(.1)
                        time += .1
                        if time >= timeout:
                            break
                else:
                    while sim_thread.is_alive():
                        sim_thread.join(.1)

                if live_grapher[0] is not None:
                    display_timer.cancel()
                self.stop_event.set()
                while self.result is None:
                    pass
            except KeyboardInterrupt:
                if live_output:
                    display_timer.pause = True
                    display_timer.cancel()
                self.pause_event.set()
                while self.result is None:
                    pass
            if hasattr(self, 'has_raised_exception'):
                raise self.has_raised_exception

            return Results.build_from_solver_results(self, live_output_options)

    def ___run(self, model, curr_state,total_time, timeline, trajectory_base, tmpSpecies, live_grapher, t=20,
               number_of_trajectories=1, increment=0.05, seed=None, debug=False, profile=False, show_labels=True,
               timeout=None, resume=None, tau_tol=0.03, **kwargs):

        try:
            self.__run(model, curr_state, total_time, timeline, trajectory_base, tmpSpecies, live_grapher, t, number_of_trajectories,
                       increment, seed, debug, profile, timeout, resume, tau_tol, **kwargs)

        except Exception as e:
            self.has_raised_exception = e
            self.result = []
            return [], -1

    def __run(self, model, curr_state, total_time, timeline, trajectory_base, tmpSpecies, live_grapher, t=20,
              number_of_trajectories=1, increment=0.05, seed=None, debug=False, profile=False, timeout=None,
              resume=None, tau_tol=0.03, **kwargs):

        # for use with resume, determines how much excess data to cut off due to
        # how species and time are initialized to 0
        timeStopped = 0
        if resume is not None:
            if resume[0].model != model:
                raise ModelError('When resuming, one must not alter the model being resumed.')
            if t < resume['time'][-1]:
                raise ExecutionError(
                    "'t' must be greater than previous simulations end time, or set in the run() method as the "
                    "simulations next end time")
        if debug:
            print("t = ", t)
            print("increment = ", increment)

        species_mappings, species, parameter_mappings, number_species = nputils.numpy_initialization(model)

        if seed is not None:
            if not isinstance(seed, int):
                seed = int(seed)
            if seed > 0:
                random.seed(seed)
                np.random.seed(seed)
            else:
                raise ModelError('seed must be a positive integer')

        simulation_data = []

        for trajectory_num in range(number_of_trajectories):
            if self.stop_event.is_set():
                self.rc = 33
                break
            elif self.pause_event.is_set():
                timeStopped = timeline[entry_count]
            # For multi trajectories, live_grapher needs to be informed of trajectory increment
            if live_grapher[0] is not None:
                live_grapher[0].increment_trajectory(trajectory_num)

            start_state = [0] * (len(model.listOfReactions) + len(model.listOfRateRules))
            propensities = {}
            curr_state[0] = {}

            if resume is not None:
                save_time = total_time[0]
                curr_time = [save_time]
            else:
                save_time = 0
                curr_time = [0]

            curr_state[0]['vol'] = model.volume
            data = {'time': timeline}
            steps_taken = []
            steps_rejected = 0
            entry_count = 0
            trajectory = trajectory_base[trajectory_num]

            HOR, reactants, mu_i, sigma_i, g_i, epsilon_i, critical_threshold = Tau.initialize(model, tau_tol)

            if resume is not None:
                for spec in model.listOfSpecies:
                    curr_state[0][spec] = tmpSpecies[spec]
            else:
                for spec in model.listOfSpecies:
                    curr_state[0][spec] = model.listOfSpecies[spec].initial_value

            for param in model.listOfParameters:
                curr_state[0][param] = model.listOfParameters[param].value

            for i, rxn in enumerate(model.listOfReactions):
                # set reactions to uniform random number and add to start_state
                start_state[i] = (math.log(random.uniform(0, 1)))
                if debug:
                    print("Setting Random number ",
                          start_state[i], " for ", model.listOfReactions[rxn].name)

            compiled_propensities = {}
            for i, r in enumerate(model.listOfReactions):
                compiled_propensities[r] = compile(model.listOfReactions[r].propensity_function, '<string>', 'eval')

            timestep = 0
            
            # Each save step
            while entry_count < timeline.size:
                if self.stop_event.is_set():
                    self.rc = 33
                    break
                elif self.pause_event.is_set():
                    timeStopped = timeline[entry_count]
                    break

                # Until save step reached
                while curr_time[0] < save_time:
                    if self.stop_event.is_set():
                        self.rc = 33
                        break
                    elif self.pause_event.is_set():
                        timeStopped = timeline[entry_count]
                        break

                    propensity_sum = 0

                    for i, r in enumerate(model.listOfReactions):
                        propensities[r] = eval(compiled_propensities[r], curr_state[0])
                        propensity_sum += propensities[r]

                    tau_args = [HOR, reactants, mu_i, sigma_i, g_i, epsilon_i, tau_tol, critical_threshold,
                                model, propensities, curr_state[0], curr_time[0], save_time]

                    tau_step = Tau.select(*tau_args)

                    prev_start_state = start_state.copy()
                    prev_curr_state = curr_state[0].copy()
                    prev_curr_time = curr_time[0]
                    total_time[0] = curr_time[0]

                    loop_cnt = 0
                    while True:
                        loop_cnt += 1
                        if loop_cnt > 100:
                            raise Exception("Loop over __get_reactions() exceeded loop count")

                        reactions, curr_state[0], curr_time[0] = self.__get_reactions(
                            tau_step, curr_state[0], curr_time[0], save_time,
                            propensities, model.listOfReactions)

                        # Update curr_state with the result of the SSA reaction that fired
                        species_modified = {}
                        for i, rxn in enumerate(model.listOfReactions):
                            if reactions[rxn] > 0:
                                for reactant in model.listOfReactions[rxn].reactants:
                                    species_modified[reactant.name] = True
                                    curr_state[0][reactant.name] -= model.listOfReactions[
                                        rxn].reactants[reactant] * reactions[rxn]
                                for product in model.listOfReactions[rxn].products:
                                    species_modified[product.name] = True
                                    curr_state[0][product.name] += model.listOfReactions[
                                        rxn].products[product] * reactions[rxn]
                        neg_state = False
                        for spec in species_modified:
                            if curr_state[0][spec] < 0:
                                neg_state = True
                                if debug:
                                    print("Negative state detected: curr_state[{0}]= {1}".format(spec,
                                                                                                 curr_state[0][spec]))
                        if neg_state:
                            if debug:
                                print("\trxn={0}".format(reactions))
                            start_state = prev_start_state.copy()
                            curr_state[0] = prev_curr_state.copy()
                            curr_time[0] = prev_curr_time
                            total_time[0] = prev_curr_time

                            tau_step = tau_step / 2
                            steps_rejected += 1
                            if debug:
                                print("Resetting curr_state[{0}]= {1}".format(spec, curr_state[0][spec]))
                            if debug:
                                print("\tRejecting step, taking step of half size, tau_step={0}".format(tau_step))
                        else:
                            break  # breakout of the while True

                # save step reached
                for i in range(number_species):
                    trajectory[entry_count][i + 1] = curr_state[0][species[i]]
                save_time += increment
                timestep += 1
                entry_count += 1
                
            # end of trajectory
            for i in range(number_species):
                data[species[i]] = trajectory[:, i+1]
            simulation_data.append(data)

            if profile:
                print(steps_taken)
                print("Total Steps Taken: ", len(steps_taken))
                print("Total Steps Rejected: ", steps_rejected)

        # If simulation has been paused, or tstopped !=0
        if timeStopped != 0 or resume is not None:
            simulation_data = nputils.numpy_resume(timeStopped, simulation_data, resume=resume)

        self.result = simulation_data
        return self.result, self.rc



"""
GillesPy2 is a modeling toolkit for biochemical simulation.
Copyright (C) 2019-2021 GillesPy2 developers.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

"""
This Python module contains the initialization and selection methods for the Tau-Leaping method described in Cao, Y.;
Gillespie, D. T.; Petzold, L. R. (2006). "Efficient step size selection for the tau-leaping simulation method" (PDF).
The Journal of Chemical Physics. 124 (4): 044109. Bibcode:2006JChPh.124d4109C. doi:10.1063/1.2159468. PMID 16460151.
This module is for use in the basic_tau_leaping_solver and basic_tau_hybrid solver only.
"""


def initialize(model, epsilon):
    """
    This method initailizes the state for tau-leaping selections to be made.
    Based on Cao, Y.; Gillespie, D. T.; Petzold, L. R. (2006). "Efficient step size selection for the tau-leaping
    simulation method" (PDF).
    The Journal of Chemical Physics. 124 (4): 044109. Bibcode:2006JChPh.124d4109C. doi:10.1063/1.2159468. PMID 16460151
    """

    HOR = {}  # Highest Order Reaction of species
    reactants = set()  # a list of all species in the model which act as reactants
    mu_i = {}  # mu_i for each species
    sigma_i = {}  # sigma_i squared for each species
    g_i = {}  # Relative species error allowance denominator
    epsilon_i = {}  # Relative error allowance of species
    critical_threshold = 10  # Reactant Population to be considered critical

    # initialize Highest Order Reactions
    for s in model.listOfSpecies:
        HOR[s] = 0

    # Create list of all reactants
    for r in model.listOfReactions:
        # Calculate this reaction's order
        reaction_order = sum(model.listOfReactions[r].reactants.values())
        for reactant, count in model.listOfReactions[r].reactants.items():
            # Build reactant list
            reactants.add(reactant)
            # Initialize mu and sigma for each reactant
            mu_i[reactant.name] = 0
            sigma_i[reactant.name] = 0
            # if this reaction's order is higher than previous, set HOR
            if reaction_order > HOR[reactant.name]:
                HOR[reactant.name] = reaction_order
                if count == 2 and reaction_order == 2:
                    g_i[reactant.name] = lambda x: 2 + (1 / (x - 1))
                elif count == 2 and reaction_order == 3:
                    g_i[reactant.name] = lambda x: (3 / 2) * (2 + (1 / (x - 1)))
                elif count == 3:
                    g_i[reactant.name] = lambda x: (3 + (1 / (x - 1)) + (2 / (x - 2)))
                else:
                    g_i[reactant.name] = HOR[reactant.name]
                    epsilon_i[reactant.name] = epsilon / g_i[reactant.name]

    # Return components for tau selection
    return HOR, reactants, mu_i, sigma_i, g_i, epsilon_i, critical_threshold


def select(*tau_args):
    """
    Tau Selection method based on Cao, Y.; Gillespie, D. T.; Petzold, L. R. (2006).
    "Efficient step size selection for the tau-leaping simulation method" (PDF).
    The Journal of Chemical Physics. 124 (4): 044109. Bibcode:2006JChPh.124d4109C. doi:10.1063/1.2159468. PMID 16460151
    """

    HOR, reactants, mu_i, sigma_i, g_i, epsilon_i, epsilon, critical_threshold, model, propensities, curr_state, \
    curr_time, save_time = tau_args
    tau_step = 0
    crit_taus = {}  # Estimated time to single-firing of critical reactions
    critical_reactions = []  # List of critical reactions at this step
    critical = False  # system-wide flag, true when any reaction is critical
    critical_tau = 0  # holds the smallest tau time for critical reactions
    non_critical_tau = 0  # holds the smallest tau time for non-critical reactions
    tau = 0

    # Determine if there are any critical reactions
    for rxn in model.listOfReactions:
        for r, v in model.listOfReactions[rxn].reactants.items():
            if curr_state[r.name] / v < critical_threshold and propensities[rxn] > 0:
                critical_reactions.append(rxn)
                critical = True

    # If a critical reaction is present, estimate tau for a single firing of each
    # critical reaction with propensity > 0, and take the smallest tau
    if critical:
        for rxn in model.listOfReactions:
            if propensities[rxn] > 0:
                crit_taus[rxn] = 1 / propensities[rxn]
        critical_tau = min(crit_taus.values())

    # If a reactant's HOR requires >1 of that reactant, evaluate lambda at curr_state
    for r in g_i:
        if callable(g_i[r]):
            g_i[r] = g_i[r](curr_state[r])
            epsilon_i[r] = epsilon / g_i[r]

    tau_i = {}  # estimated tau for non-critical reactions
    mu_i = {species.name: 0 for species in model.listOfSpecies.values()}
    sigma_i = {species.name: 0 for species in model.listOfSpecies.values()}

    for r in model.listOfReactions:
        # Calculate abs mean and standard deviation for each reactant
        for reactant in model.listOfReactions[r].reactants:
            if reactant not in mu_i:
                mu_i[reactant.name] = 0
            if reactant not in sigma_i:
                sigma_i[reactant.name] = 0
            mu_i[reactant.name] += model.listOfReactions[r].reactants[reactant] * propensities[
                r]  # Cao, Gillespie, Petzold 32a
            sigma_i[reactant.name] += model.listOfReactions[r].reactants[reactant] ** 2 * propensities[
                r]  # Cao, Gillespie, Petzold 32b

    for r in reactants:
        calculated_max = epsilon_i[r.name] * curr_state[r.name]
        max_pop_change_mean = max(calculated_max, 1)
        max_pop_change_sd = max(calculated_max, 1) ** 2
        if mu_i[r.name] > 0:
            # Cao, Gillespie, Petzold 33
            tau_i[r.name] = min(
                abs(max_pop_change_mean / mu_i[r.name]),
                max_pop_change_sd / sigma_i[r.name])

    if len(tau_i) > 0: non_critical_tau = min(tau_i.values())

    for r in model.listOfReactions:
        # Calculate abs mean and standard deviation for each reactant
        for product in model.listOfReactions[r].products:
            mu_i[product.name] -= model.listOfReactions[r].products[product] * propensities[
                r]  # Cao, Gillespie, Petzold 32a
            sigma_i[product.name] += model.listOfReactions[r].products[product] ** 2 * propensities[
                r]  # Cao, Gillespie, Petzold 32b

    # If all reactions are non-critical, use non-critical tau.
    if not critical:
        tau = non_critical_tau
    # If all rxns are critical, use critical tau.
    elif len(tau_i) == 0:
        tau = critical_tau
    # If there are both critical and non-critical reactions,
    # take the shortest tau between critical and non-critical.
    else:
        tau = min(non_critical_tau, critical_tau)
    # If selected tau exceeds save time, integrate to save time
    if tau > 0:
        tau = max(tau, 1e-10)  # set minimum to prevent integration errors
        if save_time - curr_time > 0:
            tau = min(tau, save_time - curr_time)
    else:
        tau = save_time - curr_time
    return tau



from datetime import time
import unittest
import numpy as np
import io
from gillespy2.solvers import ODECSolver, SSACSolver, TauLeapingCSolver, TauHybridCSolver
from example_models import Degradation
from gillespy2.solvers.cpp.c_decoder import BasicSimDecoder, IterativeSimDecoder

class TestCDecode(unittest.TestCase):
    model = Degradation()
    solvers = [
        # Solver     List of trajectory numbers to try
        (ODECSolver,        [1]),
        (SSACSolver,        [1, 3]),
        (TauLeapingCSolver, [1, 3]),
        (TauHybridCSolver,  [1, 3]),
    ]
    decoders = [BasicSimDecoder, IterativeSimDecoder]

    def setUp(self):
        self.model.timespan(np.linspace(0, 300, 301))
        self.solvers = [(solver(model=self.model), trajectory_numbers) for solver, trajectory_numbers in self.solvers]

    def test_run_output(self):
        expected_tspan = self.model.tspan
        expected_init = self.model.get_species("A").initial_value
        for solver, trajectory_counts in self.solvers:
            for number_of_trajectories in trajectory_counts:
                with self.subTest("Processing simulation output for each solver with different trajectory sizes",
                                  number_of_trajectories=number_of_trajectories,
                                  solver=solver):
                    results = self.model.run(solver=solver, number_of_trajectories=number_of_trajectories, seed=1024)
                    for result in results:
                        tspan, first_value, last_value = result["time"], result["A"][0], result["A"][-1]
                        result_diff = np.concatenate([np.array([first_value]), result["A"][1:]])
                        result_diff = result["A"] - result_diff
                        self.assertTrue(np.allclose(tspan, expected_tspan), msg="C++ Simulation output contains unexpected timeline values"
                                                                        f"\n  Received timeline: {tspan}\n  Expected timeline: {expected_tspan}")
                        self.assertEqual(first_value, expected_init, msg=f"C++ Simulation output begins with unexpected value: {first_value}")
                        self.assertAlmostEqual(last_value, 0, places=3, msg=f"C++ Simulation output converges on an unexpectedly large value: {last_value}")

    def test_c_decoder(self):
        """
        Ensures that, given a certain input from stdout, the results will be
            properly formatted by the sim decoders.
        """
        example_input = """
        0,1.0,2.0,3.0,1,2.0,4.0,6.0,2,3.0,6.0,9.0,3,4.0,8.0,12.0,
        0,1.0,2.0,3.0,1,2.0,4.0,6.0,2,3.0,6.0,9.0,3,4.0,8.0,12.0,3
        """.strip()
        expected_result = np.array([
            [ [0, 1.0,2.0,3.0], [1, 2.0,4.0,6.0], [2, 3.0,6.0,9.0], [3, 4.0,8.0,12.0] ],
            [ [0, 1.0,2.0,3.0], [1, 2.0,4.0,6.0], [2, 3.0,6.0,9.0], [3, 4.0,8.0,12.0] ]
        ])

        for decoder in TestCDecode.decoders:
            mock_stdout = io.BytesIO(bytes(example_input, "ascii"))
            mock_stdout = io.BufferedReader(mock_stdout)
            with self.subTest("Processing mock data output with C decoders", decoder=decoder):
                reader = decoder(np.zeros((2, 4, 4)))
                reader.read(mock_stdout)
                result, time_stopped = reader.get_output()

                self.assertEqual(time_stopped, 3, f"Stop time returned by decoder is incorrect (expected {3}, got {time_stopped})")
                # Test both the passed-in trajectory list and the returned trajectories.
                self.assertTrue(np.all(result == expected_result), "Trajectory data returned by decoder is incorrect")



"""
GillesPy2 is a modeling toolkit for biochemical simulation.
Copyright (C) 2019-2021 GillesPy2 developers.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import random, math, sys, warnings
from collections import OrderedDict
from scipy.integrate import ode, LSODA
import heapq
import numpy as np
import threading
import gillespy2
from gillespy2.solvers.utilities import Tau
from gillespy2.core import GillesPySolver, log
from gillespy2.core.gillespyError import *
from gillespy2.core.results import Results

eval_globals = math.__dict__

def __piecewise(*args):
    # Eval entry for piecewise functions
    args = list(args)
    sol = None
    if len(args) % 2:
        args.append(True)
    for i, arg in enumerate(args):
        if not i % 2:
            continue
        if arg:
            sol = args[i - 1]
            break
    return sol


def __xor(*args):
    # Eval entry for MathML xor function
    from operator import ixor
    from functools import reduce
    args = list(args)
    return reduce(ixor, args)


eval_globals['false'] = False
eval_globals['true'] = True
eval_globals['piecewise'] = __piecewise
eval_globals['xor'] = __xor


class TauHybridSolver(GillesPySolver):
    """
    This Solver uses a root-finding interpretation of the direct SSA method,
    along with ODE solvers to simulate ODE and Stochastic systems
    interchangeably or simultaneously.
    """
    name = "TauHybridSolver"
    rc = 0
    result = None
    stop_event = None

    def __init__(self, model=None):
        name = 'TauHybridSolver'
        rc = 0
        self.model = model

    def __toggle_reactions(self, model, all_compiled, deterministic_reactions, dependencies, 
                            curr_state, det_spec, rr_sets):
        """
        Helper method which is used to convert reaction channels into
        rate rules, and rate rules into reaction channels, as they are switched
        dynamically throughout the simulation based on user-supplied tolerance.
        """

        # initialize variables
        inactive_reactions = all_compiled['inactive_rxns']
        rate_rules = all_compiled['rules']
        rxns = all_compiled['rxns']


        # If the set has changed, reactivate non-determinsitic reactions
        reactivate = []
        for r in inactive_reactions:
            if not r in deterministic_reactions:
                reactivate.append(r)
        for r in reactivate:
            rxns[r] = inactive_reactions.pop(r, None)

        # floor non-det species
        for s, d in det_spec.items():
            if not d and isinstance(curr_state[s], float):
                curr_state[s] = math.floor(curr_state[s])

        # Deactivate Determinsitic Reactions
        for r in deterministic_reactions:
            if not r in inactive_reactions:
                inactive_reactions[r] = rxns.pop(r, None)

        # Check if this reaction set is already compiled and in use:
        if deterministic_reactions in rr_sets.keys():
            return rr_sets[deterministic_reactions]
        else:
        # Otherwise, this is a new determinstic reaction set that must be compiled
            return self.__create_diff_eqs(deterministic_reactions, model,
                                            dependencies, rr_sets)

    def __create_diff_eqs(self, comb, model, dependencies, rr_sets):
        """
        Helper method used to convert stochastic reaction descriptions into
        differential equations, used dynamically throught the simulation.
        """
        diff_eqs = OrderedDict()
        rate_rules = OrderedDict()

        # Initialize sample dict
        rr_vars = {}
        for n, rr in model.listOfRateRules.items():
            rr_vars[rr.variable] = n
        for spec in model.listOfSpecies:
            if spec in rr_vars.keys():
                diff_eqs[model.listOfSpecies[spec]] = model.listOfRateRules[rr_vars[spec]].formula
            else:
                diff_eqs[model.listOfSpecies[spec]] = '0'

        # loop through each det reaction and concatenate it's diff eq for each species
        for reaction in comb:
            factor = {dep: 0 for dep in dependencies[reaction]}

            for key, value in model.listOfReactions[reaction].reactants.items():
                if not key.constant and not key.boundary_condition:
                    factor[key.name] -= value
            for key, value in model.listOfReactions[reaction].products.items():
                if not key.constant and not key.boundary_condition:
                    factor[key.name] += value

            for dep in dependencies[reaction]:
                if factor[dep] != 0:
                    if model.listOfSpecies[dep].mode == 'continuous':
                        diff_eqs[model.listOfSpecies[dep]] += ' + {0}*({1})'.format(factor[dep],
                                                               model.listOfReactions[reaction].ode_propensity_function)
                    else:
                        diff_eqs[model.listOfSpecies[dep]] += ' + {0}*({1})'.format(factor[dep],
                                                               model.listOfReactions[reaction].propensity_function)

        for spec in model.listOfSpecies:
            if diff_eqs[model.listOfSpecies[spec]] == '0':
                del diff_eqs[model.listOfSpecies[spec]]
        # create a dictionary of compiled gillespy2 rate rules
        for spec, rate in diff_eqs.items():
            rate_rules[spec] = compile(gillespy2.RateRule(spec, rate).formula, '<string>', 'eval')
        rr_sets[comb] = rate_rules # save values
        return rate_rules

    def __flag_det_reactions(self, model, det_spec, det_rxn, dependencies):
        """
        Helper method used to flag reactions that can be processed
        deterministically without exceeding the user-supplied tolerance.
        """
        # Determine if each rxn would be deterministic apart from other reactions
        prev_state = det_rxn.copy()
        for rxn in model.listOfReactions:
            # assume it is deterministic
            det_rxn[rxn] = True
            # iterate through the dependent species of this reaction
            for species in dependencies[rxn]:
                # if any of the dependencies are discrete or (dynamic AND the 
                # species itself has not been flagged as deterministic)
                # then allow it to be modelled discretely
                if model.listOfSpecies[species].mode == 'discrete':
                    det_rxn[rxn] = False
                    break
                if model.listOfSpecies[species].mode == 'dynamic' and det_spec[species] == False:
                    det_rxn[rxn] = False
                    break

        # Create a hashable frozenset of all determinstic reactions
        deterministic_reactions = set()
        for rxn in det_rxn:
            if det_rxn[rxn]:
                deterministic_reactions.add(rxn)
        deterministic_reactions = frozenset(deterministic_reactions)
        return deterministic_reactions

    def __calculate_statistics(self, *switch_args):
        """
        Calculates Mean, Standard Deviation, and Coefficient of Variance for each
        dynamic species, then set if species can be represented determistically
        """
        model, propensities, curr_state, tau_step, det_spec = switch_args

        CV = OrderedDict()
        mn = {species: curr_state[species] for species, value in
              model.listOfSpecies.items() if value.mode == 'dynamic'}
        sd = {species: 0 for species, value in
              model.listOfSpecies.items() if value.mode == 'dynamic'}

        for r, rxn in model.listOfReactions.items():
            for reactant in rxn.reactants:
                if reactant.mode == 'dynamic':
                    mn[reactant.name] -= (tau_step * propensities[r] * rxn.reactants[reactant])
                    sd[reactant.name] += (tau_step * propensities[r] * rxn.reactants[reactant] ** 2)
            for product in rxn.products:
                if product.mode == 'dynamic':
                    mn[product.name] += (tau_step * propensities[r] * rxn.products[product])
                    sd[product.name] += (tau_step * propensities[r] * rxn.products[product] ** 2)

        # Get coefficient of variance for each dynamic species
        for species in mn:
            sref = model.listOfSpecies[species]
            if sref.switch_min == 0:
                if mn[species] > 0:
                    CV[species] = sd[species] / mn[species]
                else:
                    CV[species] = 1  # value chosen to guarantee discrete
                # Set species to deterministic if CV is less than threshhold
                det_spec[species] = CV[species] < sref.switch_tol
            else:
                det_spec[species] = mn[species] > sref.switch_min
        return sd, CV

    @staticmethod
    def __f(t, y, curr_state, species, reactions, rate_rules, propensities,
            y_map, compiled_reactions, active_rr, events, assignment_rules):
        """
        Evaluate the propensities for the reactions and the RHS of the Reactions and RateRules.
        Also evaluates boolean value of event triggers.
        """
        state_change = [0] * len(y_map)
        curr_state['t'] = t
        curr_state['time'] = t
        for item, index in y_map.items():
            if item in assignment_rules:
                curr_state[assignment_rules[item].variable] = eval(assignment_rules[item].formula,
                                                                   {**eval_globals, **curr_state})
            else:
                curr_state[item] = y[index]
        for s, rr in active_rr.items():
            try:
                state_change[y_map[s.name]] += eval(rr, {**eval_globals, **curr_state})
            except ValueError:
                pass
        for i, r in enumerate(compiled_reactions):
            propensities[r] = eval(compiled_reactions[r], {**eval_globals, **curr_state})
            state_change[y_map[r]] += propensities[r]
        for event in events:
            triggered = eval(event.trigger.expression, {**eval_globals, **curr_state})
            if triggered: 
                state_change[y_map[event]] = 1

        return state_change

    def __find_event_time(self, sol, model, start, end, index, depth):
        """
        Helper method providing binary search implementation for locating
        precise event times.
        """
        dense_range = np.linspace(start, end, 3)
        mid = dense_range[1]
        if start >= mid or mid >= end or depth == 20: return end
        solutions = np.diff(sol.sol(dense_range)[-len(model.listOfEvents) + index])
        bool_res = [x > 0 for x in solutions]

        if bool_res[0]:  # event before mid
            depth += 1
            return self.__find_event_time(sol, model, dense_range[0],
                                          dense_range[1], index, depth)
        else:  # event after mid
            depth += 1
            return self.__find_event_time(sol, model, dense_range[1],
                                          dense_range[2], index, depth)

    def __detect_events(self, event_sensitivity, sol, model, delayed_events,
                        trigger_states, curr_time, curr_state):
        """
        Helper method to locate precise time of event firing.  This method
        first searches for any instance of an event using event_sensitivity to
        determine the granularity of the search.  If an event is detected, a
        binary search is then used to locate the precise time of that event.
        """
        event_times = {}
        dense_range = np.linspace(sol.t[0], sol.t[-1], len(sol.t) * event_sensitivity)
        solutions = np.diff(sol.sol(dense_range))
        for i, e in enumerate(model.listOfEvents.values()):
            bool_res = [x > 0 for x in solutions[i - len(model.listOfEvents)]]
            curr_state[e.name] = bool_res[-1]
            # Search for changes from False to True in event, record first time
            for y in range(1, len(dense_range) - 1):
                # Check Persistent Delays.  If an event is not designated as
                # persistent, and the trigger expression fails to evaluate as
                # true before assignment is carried out, remove the event from
                # the queue.
                if e.name in trigger_states and not e.trigger.persistent:
                    if not bool_res[y]:
                        delayed_events[i] = delayed_events[-1]  # move to end
                        heapq.heappop(delayed_events)
                        del trigger_states[e.name]
                        curr_state[e.name] = False
                # IF triggered from false to true, refine search
                elif bool_res[y] and dense_range[y] != curr_time and bool_res[y - 1] == 0:
                    event_time = self.__find_event_time(sol, model, dense_range[y - 1],
                                                        dense_range[y + 1], i, 0)
                    if event_time in event_times:
                        event_times[event_time].append(e)
                    else:
                        event_times[event_time] = [e]
                    break
        return event_times

    def __get_next_step(self, event_times, reaction_times, delayed_events,
                        sim_end, next_tau):
        """
        Helper method to determine the next action to take during simulation,
        and returns that action along with the time that it occurs.
        """

        next_event_trigger = sim_end + 2
        next_delayed_event = sim_end + 3
        curr_time = sim_end

        # event triggers
        if len(event_times):
            next_event_trigger = min(event_times)

        # delayed events
        if len(delayed_events):
            next_delayed_event = delayed_events[0][0]

        # Execute rxns/events
        next_step = {sim_end: 'end', next_tau: 'tau', next_event_trigger: 'trigger', next_delayed_event: 'delay'}

        # Set time to next action
        curr_time = min(sim_end, next_tau, next_event_trigger,
                        next_delayed_event)
        return next_step[curr_time], curr_time

    def __process_queued_events(self, model, event_queue, trigger_states,
                                curr_state):
        """
        Helper method which processes the events queue. Method is primarily for
        evaluating assignments at the designated state (trigger time or event
        time).
        """
        # Process all queued events
        events_processed = []
        pre_assignment_state = curr_state.copy()
        while event_queue:
            # Get events in priority order
            fired_event = model.listOfEvents[heapq.heappop(event_queue)[1]]
            events_processed.append(fired_event)
            if fired_event.name in trigger_states:
                assignment_state = trigger_states[fired_event.name]
                del trigger_states[fired_event.name]
            else:
                assignment_state = pre_assignment_state
            for a in fired_event.assignments:
                # Get assignment value
                assign_value = eval(a.expression, eval_globals, assignment_state)
                # Update state of assignment variable
                curr_state[a.variable.name] = assign_value

        return events_processed

    def __handle_event(self, event, curr_state, curr_time, event_queue,
                       trigger_states, delayed_events):
        """
        Helper method providing logic for updating states based on assignments.
        """
        # Fire trigger time events immediately
        if event.delay is None:
            heapq.heappush(event_queue, (eval(event.priority), event.name))
        # Queue delayed events
        else:
            curr_state['t'] = curr_time
            curr_state['time'] = curr_time
            execution_time = curr_time + eval(event.delay, {**eval_globals, **curr_state})
            curr_state[event.name] = True
            heapq.heappush(delayed_events, (execution_time, event.name))
            if event.use_values_from_trigger_time:
                trigger_states[event.name] = curr_state.copy()
            else:
                trigger_states[event.name] = curr_state

    def __check_t0_events(self, model, initial_state):
        """
        Helper method for firing events who reach a trigger condition at start
        of simulation, time == 0.
        """
        # Check Event State at t==0
        species_modified_by_events = []
        t0_delayed_events = {}
        for e in model.listOfEvents.values():
            if not e.trigger.value:
                t0_firing = eval(e.trigger.expression, {**eval_globals, **initial_state})
                if t0_firing:
                    if e.delay is None:
                        for a in e.assignments:
                            initial_state[a.variable.name] = eval(a.expression, {**eval_globals, **initial_state})
                            species_modified_by_events.append(a.variable.name)
                    else:
                        execution_time = eval(e.delay, {**eval_globals, **initial_state})
                        t0_delayed_events[e.name] = execution_time
        return t0_delayed_events, species_modified_by_events

    def __update_stochastic_rxn_states(self, model, compiled_reactions, curr_state):
        """
        Helper method for updating the state of stochastic reactions.
        """
        rxn_count = OrderedDict()
        species_modified = OrderedDict()
        # Update stochastic reactions
        for rxn in compiled_reactions:
            rxn_count[rxn] = 0
            while curr_state[rxn] > 0:
                rxn_count[rxn] += 1
                curr_state[rxn] += math.log(random.uniform(0, 1))
            if rxn_count[rxn]:
                for reactant in model.listOfReactions[rxn].reactants:
                    species_modified[reactant.name] = True
                    curr_state[reactant.name] -= model.listOfReactions[rxn].reactants[reactant] * rxn_count[rxn]
                for product in model.listOfReactions[rxn].products:
                    species_modified[product.name] = True
                    curr_state[product.name] += model.listOfReactions[rxn].products[product] * rxn_count[rxn]
        return species_modified

    def __integrate(self, integrator, integrator_options, curr_state, y0, model, curr_time,
                    propensities, y_map, compiled_reactions,
                    active_rr, event_queue,
                    delayed_events, trigger_states,
                    event_sensitivity, tau_step, pure_ode):
        """ 
        Helper function to perform the ODE integration of one step.  This
        method uses scipy.integrate.LSODA to get simulation data, and
        determines the next stopping point of the simulation. The state is
        updated and returned to __simulate along with curr_time and the
        solution object. 
        """
        max_step_size = model.tspan[1] - model.tspan[0] / 100

        from functools import partial
        events = model.listOfEvents.values()
        dense_output = False 
        int_args = [curr_state, model.listOfSpecies, model.listOfReactions,
                    model.listOfRateRules,
                    propensities, y_map,
                    compiled_reactions,
                    active_rr,
                    events,
                    model.listOfAssignmentRules]
        rhs = lambda t, y: TauHybridSolver.__f(t, y, *int_args)
        if 'min_step' in integrator_options:
            tau_step = max(integrator_options['min_step'], tau_step)
        else:
            tau_step = max(1e-6, tau_step)
        if pure_ode:
            next_tau = curr_time+max_step_size
            dense_output = True
        else:
            next_tau = curr_time + tau_step
        curr_state['t'] = curr_time
        curr_state['time'] = curr_time

        # Integrate until end or tau is reached
        loop_count = 0
        sol = LSODA(rhs, curr_time, y0, next_tau)
        counter = 0
        while sol.t < next_tau:
            counter += 1
            sol.step()
            # Update states of all species based on changes made to species through
            # ODE processes.  This will update all species whose mode is set to
            # 'continuous', as well as 'dynamic' mode species which have been
            # flagged as deterministic.
            for spec_name, species in model.listOfSpecies.items():
                if not species.constant:
                    curr_state[spec_name] = sol.y[y_map[spec_name]]

        # Search for precise event times
        '''
        if len(model.listOfEvents):
            event_times = self.__detect_events(event_sensitivity, sol, model, delayed_events,
                                               trigger_states, curr_time, curr_state)
        else:
            event_times = {}

        # Get next tau time
        reaction_times = []
        '''
        # Set curr time to next time a change occurs in the system outside of
        # the standard ODE process.  Determine what kind of change this is,
        # and set the curr_time of simulation to restart simulation after
        # making the appropriate state changes.
        event_times = {}
        reaction_times = []
        next_step, curr_time = self.__get_next_step(event_times, reaction_times,
                                                    delayed_events,
                                                    model.tspan[-1], next_tau)
        curr_state['t'] = curr_time

        # Stochastic Reactions are also fired through a root-finding method
        # which mirrors the standard SSA probability.  Since we are using
        # Tau-Leaping, rather than firing reactions based on a direct poisson
        # distribution from the propensity, we use that same poisson
        # distribution to select a random negative offset for the reaction
        # state, then integrate forward along with any deterministic
        # reactions/species for a designated time (tau).  Root crossings
        # satisfy the SSA firing process, and because multiple reactions can be
        # fired in a single Tau step, a new random number will be generated and
        # added (representing a single-firing each time) until the state value
        # of the reaction is once again negative.
        for rxn in compiled_reactions:
            curr_state[rxn] = sol.y[y_map[rxn]]

        # In the case that a major change occurs to the system (outside of the
        # standard ODE process) is caused by an event trigger, we then examine
        # the event which was triggered.  if Event assignments are carried out
        # immediately, we add them to a heap in Event Priority order to be
        # immediately dealt with.  If an Event contains a delay, we can now set
        # the time of event assignment execution by evaluating the event delay
        # in the current state.
        if next_step == 'trigger':
            for event in event_times[curr_time]:
                self.__handle_event(event, curr_state, curr_time,
                                    event_queue, trigger_states, delayed_events)
        # In the case that a major change occurs to the system (outside of the
        # standard ODE process) is caused by a delayed event which has now
        # reached the designated time of execution, we add all currently
        # designated event assignments to a heap to be processed in Event
        # Priority order.
        elif next_step == 'delay':
            event = heapq.heappop(delayed_events)
            heapq.heappush(event_queue, (eval(model.listOfEvents[event[1]].priority), event[1]))

        return sol, curr_time

    def __simulate(self, integrator, integrator_options, curr_state, y0, model, curr_time,
                   propensities, species, parameters, compiled_reactions,
                   active_rr, y_map, trajectory, save_times,
                   delayed_events, trigger_states, event_sensitivity,
                   tau_step, pure_ode, debug):
        """
        Function to process simulation until next step, which can be a
        stochastic reaction firing, an event trigger or assignment, or end of
        simulation.

        :param curr_state: Contains all state variables for system at current time
        :type curr_state: dict

        :param curr_time: Represents current time
        :type curr_time: float

        :param save_times: Currently unreached save points
        :type save_times: list

        :returns: curr_state, curr_time, save_times, sol
            sol - Python object returned from LSODA which contains all solution
            data.
        """

        event_queue = []
        prev_y0 = y0.copy()
        prev_curr_state = curr_state.copy()
        prev_curr_time = curr_time
        loop_count = 0

        while True:
            loop_count += 1
            if loop_count > 100:
                raise Exception("Loop over __integrate() exceeded loop count")
            sol, curr_time = self.__integrate(integrator, integrator_options, curr_state,
                                              y0, model, curr_time, propensities, y_map,
                                              compiled_reactions,
                                              active_rr,
                                              event_queue,
                                              delayed_events,
                                              trigger_states,
                                              event_sensitivity,
                                              tau_step,
                                              pure_ode)

            species_modified = self.__update_stochastic_rxn_states(model,
                                                                   compiled_reactions, curr_state)

            # Occasionally, a tau step can result in an overly-aggressive
            # forward step and cause a species population to fall below 0,
            # which would result in an erroneous simulation.  If this occurs,
            # back simulation up one step and attempt forward simulation using
            # a smaller tau step.
            neg_state = False
            for s in species_modified.keys():
                if curr_state[s] < 0:
                    neg_state = True
            if neg_state:
                y0 = prev_y0.copy()
                curr_state = prev_curr_state.copy()
                curr_time = prev_curr_time
                tau_step = tau_step / 2
            else:
                break

            # Now update the step and trajectories for this step of the simulation.
        # Here we make our final assignments for this step, and begin
        # populating our results trajectory.
        num_saves = 0
        for time in save_times:
            if time > curr_time:
                break
            # if a solution is given for it
            trajectory_index = np.where(model.tspan == time)[0][0]
            assignment_state = curr_state.copy()
            for s in range(len(species)):
                # Get ODE Solutions
                trajectory[trajectory_index][s + 1] = sol.y[s]
                # Update Assignment Rules for all processed time points
                if len(model.listOfAssignmentRules):
                    # Copy ODE state for assignments
                    assignment_state[species[s]] = sol.y[s]
            assignment_state['t'] = time
            for ar in model.listOfAssignmentRules.values():
                assignment_value = eval(ar.formula, {**eval_globals, **assignment_state})
                assignment_state[ar.variable] = assignment_value
                trajectory[trajectory_index][species.index(ar.variable) + 1] = assignment_value
            num_saves += 1
        save_times = save_times[num_saves:]  # remove completed save times

        events_processed = self.__process_queued_events(model, event_queue, trigger_states, curr_state)

        # Finally, perform a final check on events after all non-ODE assignment
        # changes have been carried out on model.
        event_cycle = True
        while event_cycle:
            event_cycle = False
            for i, e in enumerate(model.listOfEvents.values()):
                triggered = eval(e.trigger.expression, {**eval_globals, **curr_state})
                if triggered and not curr_state[e.name]:
                    curr_state[e.name] = True
                    self.__handle_event(e, curr_state, curr_time,
                                        event_queue, trigger_states, delayed_events)
                    event_cycle = True
                elif not triggered:
                    curr_state[e.name] = False

        events_processed = self.__process_queued_events(model, event_queue, trigger_states, curr_state)

        return sol, curr_state, curr_time, save_times

    def __set_seed(self, seed):
        # Set seed if supplied
        if seed is not None:
            if not isinstance(seed, int):
                seed = int(seed)
            if seed > 0:
                random.seed(seed)
            else:
                raise ModelError('seed must be a positive integer')

    def __set_recommended_ode_defaults(self, integrator_options):
        """
        Set some ODE solver defaults.  These values are chosen based on the
        precision required to successfully complete the SBML Test suite.
        """

        if 'rtol' not in integrator_options:
            integrator_options['rtol'] = 1e-9
        if 'atol' not in integrator_options:
            integrator_options['atol'] = 1e-12
        if 'max_step' not in integrator_options:
            integrator_options['max_step'] = 0.25

    def __compile_all(self, model):
        """
        Compile all run-time evaluables to enhance performance.
        """
        compiled_reactions = OrderedDict()
        for i, r in enumerate(model.listOfReactions):
            compiled_reactions[r] = compile(model.listOfReactions[r].propensity_function, '<string>',
                                            'eval')
        compiled_rate_rules = OrderedDict()
        for i, rr in enumerate(model.listOfRateRules.values()):
            if isinstance(rr.variable, str):
                compiled_rate_rules[model.listOfSpecies[rr.variable]] = compile(
                                                            rr.formula, '<string>', 'eval')
            else:
                compiled_rate_rules[rr.variable] = compile(rr.formula, '<string>', 'eval')
        compiled_inactive_reactions = OrderedDict()

        compiled_propensities = compiled_reactions.copy()

        return compiled_reactions, compiled_rate_rules, compiled_inactive_reactions, compiled_propensities

    def __initialize_state(self, model, curr_state, debug):
        """
        Initialize curr_state for each trajectory.
        """

        # intialize parameters to current state
        for p in model.listOfParameters:
            curr_state[p] = model.listOfParameters[p].value

        # initialize species population state
        for s in model.listOfSpecies:
            curr_state[s] = model.listOfSpecies[s].initial_value

        # Set reactions to uniform random number
        for i, r in enumerate(model.listOfReactions):
            curr_state[r] = math.log(random.uniform(0, 1))
            if debug:
                print("Setting Random number ", curr_state[r], " for ", model.listOfReactions[r].name)

        # Initialize event last-fired times to 0
        for e_name in model.listOfEvents:
            curr_state[e_name] = 0

        sanitized_species = model.sanitized_species_names()
        sanitized_parameters = model.sanitized_parameter_names()
        for fd in model.listOfFunctionDefinitions.values():
            sanitized_function = fd.sanitized_function(sanitized_species, sanitized_parameters)
            curr_state[fd.name] = eval(f"lambda {', '.join(fd.args)}: {sanitized_function}", eval_globals)

        for ar in model.listOfAssignmentRules.values():
            if ar.variable in model.listOfSpecies:
                continue
            curr_state[ar.variable] = ar.formula

    def __map_state(self, species, parameters, compiled_reactions, events, curr_state):
        """
        Creates the start state vector for integration and provides a
        dictionary map to it's elements.
        """
        y_map = OrderedDict()
        # Build integration start state
        y0 = [0] * (len(species) + len(parameters) + len(compiled_reactions) + len(events))
        for i, spec in enumerate(species):
            if isinstance(curr_state[spec], str):
                y0[i] = eval(curr_state[spec], {**eval_globals, **curr_state})
            else:
                y0[i] = curr_state[spec]
            y_map[spec] = i
        for i, param in enumerate(parameters):
            y0[i + len(species)] = curr_state[param]
            y_map[param] = i + len(species)
        for i, rxn in enumerate(compiled_reactions):
            y0[i + len(species) + len(parameters)] = curr_state[rxn]
            y_map[rxn] = i + len(species) + len(parameters)
        for i, event in enumerate(events.values()):
            y0[i + len(species) + len(parameters) + len(compiled_reactions)] = curr_state[event.name]
            y_map[event] = i + len(species) + len(parameters) + len(compiled_reactions)
        return y0, y_map

    @classmethod
    def get_solver_settings(self):
        """
        :returns: Tuple of strings, denoting all keyword argument for this solvers run() method.
        """
        return ('model', 't', 'number_of_trajectories', 'increment', 'seed', 'debug', 'profile', 'tau_tol',
                'event_sensitivity', 'integrator', 'integrator_options', 'timeout')

    @classmethod
    def run(self, model=None, t=20, number_of_trajectories=1, increment=None, seed=None,
            debug=False, profile=False, tau_tol=0.03, event_sensitivity=100, integrator='LSODA',
            integrator_options={}, live_output=None, live_output_options={}, timeout=None, **kwargs):
        """
        Function calling simulation of the model. This is typically called by the run function in GillesPy2 model
        objects and will inherit those parameters which are passed with the model as the arguments this run function.

        :param model: GillesPy2 model object to simulate
        :type model: GillesPy2.model

        :param t: Simulation run time
        :type t: int

        :param number_of_trajectories: The number of times to sample the chemical master equation. Each
            trajectory will be returned at the end of the simulation.
            Optional, defaults to 1. Number of trajectories to simulate
        :type number_of_trajectories: int

        :param increment: Save point increment for recording data
        :type increment: float

        :param seed: The random seed for the simulation. Optional, defaults to None.
        :type seed: int

        :param debug: Set to True to provide additional debug information about the simulation.
        :type debug: bool

        :param profile: Set to True to provide information about step size (tau) taken at each step.
        :type profile: bool

        :param tau_tol: Tolerance level for Tau leaping algorithm.  Larger tolerance values will
            result in larger tau steps. Default value is 0.03.
        :type tau_tol: float

        :param event_sensitivity: Number of data points to be inspected between integration
            steps/save points for event detection
        :type event_sensitivity: int

        :param integrator_options:  contains options to the scipy integrator. by default, this includes
            rtol=1e-9 and atol=1e-12.  for a list of options,
            see https://docs.scipy.org/doc/scipy/reference/generated/scipy.integrate.LSODA.html.
            Example use: {max_step : 0, rtol : .01}
        :type integrator_options: dict

        :param live_output: The type of output to be displayed by solver. Can be "progress", "text", or "graph".
        :type live_output: str

        :param live_output_options: contains options for live_output. By default {"interval":1}.
            "interval" specifies seconds between displaying.
            "clear_output" specifies if display should be refreshed with each display
        :type live_output_options:  str
        """

        if isinstance(self, type):
            self = TauHybridSolver(model=model)

        increment = self.get_increment(model=model, increment=increment)

        if timeout is not None and timeout > 0:
            for i, s in enumerate(list(model._listOfSpecies.keys())):
                # Solve_ivp doesn't return any results until it's finished solving so timing out early only slows
                # the solver.
                if model.listOfSpecies[s].mode == 'continuous':
                    timeout = 0
                    log.warning('timeouts not supported by continuous species.')
                    break
                elif model.listOfSpecies[s].mode == 'dynamic':
                    log.warning('timeouts not fully supported by dynamic species. If timeout is triggered during'
                                ' integration, total solve time could be longer than expected.')
                    break

        self.stop_event = threading.Event()

        if len(kwargs) > 0:
            for key in kwargs:
                log.warning('Unsupported keyword argument to {0} solver: {1}'.format(self.name, key))

        if timeout is not None and timeout <= 0:
            timeout = None

        if debug:
            print("t = ", t)
            print("increment = ", increment)

        if len(model.listOfEvents):
            self.__set_recommended_ode_defaults(integrator_options)
        self.__set_seed(seed)

        species = list(model._listOfSpecies.keys())
        number_species = len(species)

        initial_state = OrderedDict()
        self.__initialize_state(model, initial_state, debug)
        initial_state['vol'] = model.volume
        initial_state['t'] = 0

        # create numpy array for timeline
        timeline = np.linspace(0, t, int(round(t / increment + 1)))
        model.tspan = timeline

        # create numpy matrix to mark all state data of time and species
        trajectory_base = np.zeros((number_of_trajectories, timeline.size, number_species + 1))

        # copy time values to all trajectory row starts
        trajectory_base[:, :, 0] = timeline

        # copy initial populations to base
        spec_modes = ['continuous', 'dynamic', 'discrete', None]
        for i, s in enumerate(species):
            if model.listOfSpecies[s].mode is None:
                model.listOfSpecies[s].mode = 'dynamic'

            if model.listOfSpecies[s].mode not in spec_modes:
                raise SpeciesError('Species mode can only be \'continuous\', \'dynamic\',\'discrete\', or '
                                   '\'unspecified(default to dynamic)\'.')
            trajectory_base[:, 0, i + 1] = initial_state[s]

        # curr_time and curr_state are list of len 1 so that __run receives reference
        curr_time = [0]  # Current Simulation Time
        curr_state = [None]
        live_grapher = [None]

        sim_thread = threading.Thread(target=self.___run,
                                      args=(model, curr_state, curr_time, timeline, trajectory_base, initial_state,
                                            live_grapher,), kwargs={'t': t,
                                                                    'number_of_trajectories': number_of_trajectories,
                                                                    'increment': increment, 'seed': seed,
                                                                    'debug': debug, 'profile': profile,
                                                                    'timeout': timeout, 'tau_tol': tau_tol,
                                                                    'event_sensitivity': event_sensitivity,
                                                                    'integrator': integrator,
                                                                    'integrator_options': integrator_options})
        try:
            sim_thread.start()

            if live_output is not None:
                live_output_options['type'] = live_output

                import gillespy2.core.liveGraphing
                gillespy2.core.liveGraphing.valid_graph_params(live_output_options)

                if live_output_options['type'] == "graph":
                    for i, s in enumerate(list(model._listOfSpecies.keys())):

                        if model.listOfSpecies[s].mode == 'continuous':
                            log.warning('display \"type\" = \"graph\" not recommended with continuous species. '
                                        'Try display \"type\" = \"text\" or \"progress\".')
                            break

                live_grapher[0] = gillespy2.core.liveGraphing.LiveDisplayer(model, timeline, number_of_trajectories,
                                                                            live_output_options)
                display_timer = gillespy2.core.liveGraphing.RepeatTimer(live_output_options['interval'],
                                                                        live_grapher[0].display,
                                                                        args=(curr_state, curr_time, trajectory_base, live_output))
                display_timer.start()

            sim_thread.join(timeout=timeout)

            if live_grapher[0] is not None:
                display_timer.cancel()

            self.stop_event.set()
            while self.result is None: pass
        except:
            pass
        if hasattr(self, 'has_raised_exception'):
            raise self.has_raised_exception
        
        return Results.build_from_solver_results(self, live_output_options)

    def ___run(self, model, curr_state, curr_time, timeline, trajectory_base, initial_state, live_grapher, t=20,
               number_of_trajectories=1, increment=0.05, seed=None,
               debug=False, profile=False, tau_tol=0.03, event_sensitivity=100, integrator='LSODA',
               integrator_options={}, **kwargs):
        try:
            self.__run(model, curr_state, curr_time, timeline, trajectory_base, initial_state, live_grapher, t,
                       number_of_trajectories, increment, seed, debug,
                       profile, tau_tol, event_sensitivity, integrator,
                       integrator_options, **kwargs)
        except Exception as e:
            self.has_raised_exception = e
            self.result = []
            return [], -1

    def __run(self, model, curr_state, curr_time, timeline, trajectory_base, initial_state, live_grapher, t=20,
              number_of_trajectories=1, increment=0.05, seed=None,
              debug=False, profile=False,
              tau_tol=0.03, event_sensitivity=100, integrator='LSODA',
              integrator_options={}, **kwargs):

        # create mapping of species dictionary to array indices
        species_mappings = model._listOfSpecies
        species = list(species_mappings.keys())
        parameter_mappings = model._listOfParameters
        parameters = list(parameter_mappings.keys())
        number_species = len(species)

        t0_delayed_events, species_modified_by_events = self.__check_t0_events(model, initial_state)

        # Create deterministic tracking data structures
        det_spec = {species: True for (species, value) in model.listOfSpecies.items() if value.mode == 'dynamic'}
        det_rxn = {rxn: False for (rxn, value) in model.listOfReactions.items()}

        # Determine if entire simulation is ODE or Stochastic, in order to
        # avoid unnecessary calculations during simulation
        pure_ode = True
        pure_stochastic = True
        for spec in model.listOfSpecies.values():
            if spec.mode != 'discrete':
                pure_stochastic = False
            if spec.mode != 'continuous':
                pure_ode = False

        if debug:
            print('dependencies')
            print(dependencies)

        simulation_data = []

        dependencies = OrderedDict()

        # If considering deterministic changes, create dependency data
        # structure for creating diff eqs later
        if not pure_stochastic:
            for reaction in model.listOfReactions:
                dependencies[reaction] = set()
                [dependencies[reaction].add(reactant.name) for reactant in model.listOfReactions[reaction].reactants]
                [dependencies[reaction].add(product.name) for product in model.listOfReactions[reaction].products]

        # Main trajectory loop
        for trajectory_num in range(number_of_trajectories):

            if self.stop_event.is_set():
                print('exiting')
                self.rc = 33
                break

            # For multi trajectories, live_grapher needs to be informed of trajectory increment
            if live_grapher[0] is not None:
                live_grapher[0].increment_trajectory(trajectory_num)

            trajectory = trajectory_base[trajectory_num]  # NumPy array containing this simulation's results
            propensities = OrderedDict()  # Propensities evaluated at current state

            curr_state[0] = initial_state.copy()
            curr_time[0] = 0  # Current Simulation Time

            end_time = model.tspan[-1]  # End of Simulation time
            entry_pos = 1
            data = OrderedDict()  # Dictionary for results
            data['time'] = timeline  # All time entries
            save_times = timeline

            # Record Highest Order reactant for each reaction and set error tolerance
            if not pure_ode:
                HOR, reactants, mu_i, sigma_i, g_i, epsilon_i, critical_threshold = Tau.initialize(model, tau_tol)

            # One-time compilations to reduce time spent with eval
            compiled_reactions, compiled_rate_rules, compiled_inactive_reactions, compiled_propensities = \
                self.__compile_all(model)
            all_compiled = OrderedDict()
            all_compiled['rxns'] = compiled_reactions
            all_compiled['inactive_rxns'] = compiled_inactive_reactions
            all_compiled['rules'] = compiled_rate_rules

            save_times = np.copy(model.tspan)
            delayed_events = []
            trigger_states = {}

            # Handle delayed t0 events
            for state in trigger_states.values():
                if state is None: state = curr_state[0]
            for ename, etime in t0_delayed_events.items():
                curr_state[0][ename] = True
                heapq.heappush(delayed_events, (etime, ename))
                if model.listOfEvents[ename].use_values_from_trigger_time:
                    trigger_states[ename] = curr_state[0].copy()
                else:
                    trigger_states[ename] = curr_state[0]

            # Each save step
            while curr_time[0] < model.tspan[-1]:

                if self.stop_event.is_set():
                    self.rc = 33
                    break
                # Get current propensities
                if not pure_ode:
                    for i, r in enumerate(model.listOfReactions):
                        try:
                            propensities[r] = eval(compiled_propensities[r], eval_globals, curr_state[0])
                        except Exception as e:
                            raise SimulationError('Error calculation propensity for {0}.\nReason: {1}'.format(r, e))

                # Calculate Tau statistics and select a good tau step
                if not pure_ode:
                    tau_args = [HOR, reactants, mu_i, sigma_i, g_i, epsilon_i, tau_tol, critical_threshold,
                                model, propensities, curr_state[0], curr_time[0], save_times[0]]
                tau_step = save_times[-1] - curr_time[0] if pure_ode else Tau.select(*tau_args)

                # Process switching if used
                if not pure_stochastic and not pure_ode:
                    switch_args = [model, propensities, curr_state[0], tau_step, det_spec]
                    sd, CV = self.__calculate_statistics(*switch_args)

                # Calculate sd and CV for hybrid switching and flag deterministic reactions
                if pure_stochastic:
                    deterministic_reactions = frozenset()  # Empty if non-det
                else:
                    deterministic_reactions = self.__flag_det_reactions(model, det_spec, det_rxn, dependencies)

                if debug:
                    print('mean: {0}'.format(mu_i))
                    print('standard deviation: {0}'.format(sd))
                    print('CV: {0}'.format(CV))
                    print('det_spec: {0}'.format(det_spec))
                    print('det_rxn: {0}'.format(det_rxn))

                # Set active reactions and rate rules for this integration step
                rr_sets = {frozenset() : compiled_rate_rules} # base rr set
                if pure_stochastic:
                    active_rr = rr_sets[frozenset()]
                else:
                    active_rr = self.__toggle_reactions(model, all_compiled, deterministic_reactions, 
                                                        dependencies, curr_state[0], det_spec, rr_sets)

                # Create integration initial state vector
                y0, y_map = self.__map_state(species, parameters,
                                             compiled_reactions, model.listOfEvents, curr_state[0])

                # Run simulation to next step
                sol, curr_state[0], curr_time[0], save_times = self.__simulate(integrator, integrator_options,
                                                                               curr_state[0], y0, model, curr_time[0],
                                                                               propensities, species,
                                                                               parameters, compiled_reactions,
                                                                               active_rr, y_map,
                                                                               trajectory, save_times, delayed_events,
                                                                               trigger_states,
                                                                               event_sensitivity, tau_step, pure_ode,
                                                                               debug)

            # End of trajectory, format results
            data = {'time': timeline}
            for i in range(number_species):
                data[species[i]] = trajectory[:, i + 1]
            simulation_data.append(data)

        self.result = simulation_data
        return simulation_data, self.rc



"""
GillesPy2 is a modeling toolkit for biochemical simulation.
Copyright (C) 2019-2021 GillesPy2 developers.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import unittest
from example_models import RobustModel, Example, ExampleNoTspan
from gillespy2.core import Model, Species, Reaction, Parameter
from gillespy2.core.gillespyError import *
from gillespy2.core.model import export_StochSS
import tempfile
import os
import numpy as np


class TestModel(unittest.TestCase):

    def test_model_reflexivity(self):
        model = Model()
        assert model==model

    def test_deprecate_set_parameter(self):
        model1 = Model()
        param1 = Parameter('A', expression=0)
        model1.add_parameter(param1)
        param1.set_expression(1) #raised error

    def test_model_inequality(self):
        model1 = Model()
        model2 = Model()
        param1 = Parameter('A', expression=0)
        param2 = Parameter('B', expression=1)
        model1.add_parameter(param1)
        model2.add_parameter(param2)
        assert model1 != model2

    def test_model_equality(self):
        model1 = Model()
        model2 = Model()
        param1 = Parameter('A', expression=0)
        model1.add_parameter(param1)
        model2.add_parameter(param1)
        assert model1 == model2

    def test_model_reordered_equality(self):
        model1 = Model()
        model2 = Model()
        param1 = Parameter('A', expression=0)
        param2 = Parameter('B', expression=1)
        model1.add_parameter(param1)
        model1.add_parameter(param2)
        model2.add_parameter(param2)
        model2.add_parameter(param1)
        assert model1 == model2

    def test_uniform_timespan(self):
        model = Model()
        model.timespan(np.linspace(0, 1, 100))
        with self.assertRaises(InvalidModelError):
            model.timespan(np.array([0, 0.1, 0.5]))

    def test_duplicate_parameter_names(self):
        model = Model()
        param1 = Parameter('A', expression=0)
        param2 = Parameter('A', expression=0)
        model.add_parameter(param1)
        with self.assertRaises(ModelError):
            model.add_parameter(param2)

    def test_duplicate_species_names(self):
        model = Model()
        species1 = Species('A', initial_value=0)
        species2 = Species('A', initial_value=0)
        model.add_species(species1)
        with self.assertRaises(ModelError):
            model.add_species(species2)

    def test_no_reaction_name(self):
        model = Model()
        rate = Parameter(name='rate', expression=0.5)
        model.add_parameter(rate)
        species1 = Species('A', initial_value=0)
        species2 = Species('B', initial_value=0)
        model.add_species([species1, species2])
        # add two reactions that has no name
        reaction1 = Reaction(reactants={species1: 1}, products={species2: 1}, rate=rate)
        reaction2 = Reaction(reactants={species2: 1}, products={species1: 1}, rate=rate)
        model.add_reaction([reaction1, reaction2])


#    def test_int_type_mismatch(self):
#        model = Model()
#        y1 = np.int64(5)
#        y2 = np.int32(5)
#        species1 = Species('A', initial_value=y1)
#        species2 = Species('B', initial_value=y2)
#        model.add_species([species1, species2])

    def test_duplicate_reaction_name(self):
        model = Model()
        rate = Parameter(name='rate', expression=0.5)
        model.add_parameter(rate)
        species1 = Species('A', initial_value=0)
        species2 = Species('B', initial_value=0)
        model.add_species([species1, species2])
        reaction1 = Reaction(name="reaction1", reactants={species1: 1}, products={species2: 1}, rate=rate)
        reaction2 = Reaction(name="reaction1", reactants={species2: 1}, products={species1: 1}, rate=rate)
        model.add_reaction(reaction1)
        with self.assertRaises(ModelError):
            model.add_reaction(reaction2)

    def test_model_run_with_both_increment_and_timespan(self):
        model = Example()

        try:
            model.run(increment=4)

        except SimulationError as e:
            return

        self.fail(
            """
            Failed while testing Model.run() behavior when both `timespan` and `increment` are set.
            """
        )

    def test_valid_initial_value_float(self):
        species = Species('A', initial_value=1.5, mode='continuous')

    def test_invalid_initial_value_float(self):
        with self.assertRaises(ValueError):
            species = Species('A', initial_value=1.5)

    def test_valid_initial_value_negative(self):
        species = Species('A', initial_value=-1, allow_negative_populations=True)

    def test_invalid_initial_value_negative(self):
        with self.assertRaises(ValueError):
            species = Species('A', initial_value=-1)

    def test_reaction_invalid_reactant(self):
        model = Model()
        rate = Parameter(name='rate', expression=0.5)
        model.add_parameter(rate)
        species1 = Species('A', initial_value=0)
        species2 = Species('B', initial_value=0)
        model.add_species([species1, species2])
        reaction1 = Reaction(name="reaction1", reactants={'species1': 1}, products={species2: 1}, rate=rate)
        with self.assertRaises(ModelError):
            model.add_reaction(reaction1)

    def test_reaction_invalid_product(self):
        model = Model()
        rate = Parameter(name='rate', expression=0.5)
        model.add_parameter(rate)
        species1 = Species('A', initial_value=0)
        species2 = Species('B', initial_value=0)
        model.add_species([species1, species2])
        reaction1 = Reaction(name="reaction1", reactants={species1: 1}, products={'species2': 1}, rate=rate)
        with self.assertRaises(ModelError):
            model.add_reaction(reaction1)
            
    def test_reaction_valid_reactant(self):
        model = Model()
        rate = Parameter(name='rate', expression=0.5)
        model.add_parameter(rate)
        species1 = Species('A', initial_value=0)
        species2 = Species('B', initial_value=0)
        model.add_species([species1, species2])
        reaction1 = Reaction(name="reaction1", reactants={'A': 1}, products={species2: 1}, rate=rate)
        model.add_reaction(reaction1)
        assert "reaction1" in model.listOfReactions

    def test_reaction_valid_product(self):
        model = Model()
        rate = Parameter(name='rate', expression=0.5)
        model.add_parameter(rate)
        species1 = Species('A', initial_value=0)
        species2 = Species('B', initial_value=0)
        model.add_species([species1, species2])
        reaction1 = Reaction(name="reaction1", reactants={species1: 1}, products={'B': 1}, rate=rate)
        model.add_reaction(reaction1)
        assert "reaction1" in model.listOfReactions

    def test_add_reaction_dict(self):
        model = Model()
        rate = Parameter(name='rate', expression=0.5)
        model.add_parameter(rate)
        species1 = Species('A', initial_value=0)
        species2 = Species('B', initial_value=0)
        model.add_species([species1, species2])
        reactions = {
                name: Reaction(
                    name=name, 
                    reactants={species1: 1}, products={species2: 1}, rate=rate)
                for name in ["reaction1", "reaction2"]}
        with self.assertRaises(ModelError):
            model.add_reaction(reactions)
        
    def test_species_parameter_name_substrings(self):
        model = Model()
        rate = Parameter(name='rate', expression=1)
        model.add_parameter(rate)
        species1 = Species('A', initial_value=100)
        species2 = Species('AA', initial_value=0)
        model.add_species([species1, species2])
        reaction1 = Reaction(name="reaction1", reactants={species1: 1}, products={species2: 1}, rate=rate)
        model.add_reaction(reaction1)
        number_points = 11
        model.timespan(np.linspace(0, 1, number_points))
        from gillespy2.solvers.numpy.ssa_solver import NumPySSASolver
        results = model.run(number_of_trajectories=1, solver=NumPySSASolver, seed=1)
        self.assertTrue(len(results['time']) == number_points)
        self.assertTrue(len(results[species1.name]) == number_points)
        self.assertTrue(len(results[species2.name]) == number_points)
        self.assertGreater(results[species1.name][0], results[species1.name][-1])
        self.assertLess(results[species2.name][0], results[species2.name][-1])
        self.assertEqual(np.sum(results[species1.name]) + np.sum(results[species2.name]), number_points * species1.initial_value)

    def test_problem_with_name(self):
        model = Model()
        numeric1 = Parameter(name = '123', expression = 0.5)
        with self.assertRaises(ModelError):
            model.add_parameter(numeric1)
        special1 = Parameter(name = 'parameter.', expression = 0.5)
        with self.assertRaises(ModelError):
            model.add_parameter(special1)
        special2 = Parameter(name = 'parameter[', expression = 0.5)
        with self.assertRaises(ModelError):
            model.add_parameter(special2)
        special3 = Parameter(name = 'parameter]', expression = 0.5)
        with self.assertRaises(ModelError):
            model.add_parameter(special3)
        special4 = Parameter(name = 'parameter+', expression = 0.5)
        with self.assertRaises(ModelError):
            model.add_parameter(special4)
        special5 = Parameter(name = 'parameter-', expression = 0.5)
        with self.assertRaises(ModelError):
            model.add_parameter(special5)
        special6 = Parameter(name = 'parameter*', expression = 0.5)
        with self.assertRaises(ModelError):
            model.add_parameter(special6)
        special7 = Parameter(name = 'parameter/', expression = 0.5)
        with self.assertRaises(ModelError):
            model.add_parameter(special7)
        special8 = Parameter(name = 'parameter^', expression = 0.5)
        with self.assertRaises(ModelError):
            model.add_parameter(special8)
        reserved1 = Parameter(name = "vol", expression = 0.5)
        with self.assertRaises(ModelError):
            model.add_parameter(reserved1)
        #Add this to name checks, python naming convention
        #numeric2 = Parameter(name = '1parameter', expression = 0.5)
        #with self.assertRaises(ModelError):
            #model.add_parameter(numeric2)

    def test_add_nonspecies_nonreaction_nonparameter(self):
        model = Model()
        species = 'nonspecies'
        with self.assertRaises(ModelError):
            model.add_species(species)
        reaction = 'nonreaction'
        with self.assertRaises(ModelError):
            model.add_reaction(reaction)
        parameter = 'nonparameter'
        with self.assertRaises(ParameterError):
            model.add_parameter(parameter)
        
    def test_add_event(self):
        from gillespy2.core.events import Event, EventTrigger, EventAssignment
        model = Model()
        model.add_species(Species(name="A", initial_value=1, mode="discrete"))
        model.add_species(Species(name="B", initial_value=2, mode="discrete"))
        e1t = EventTrigger(expression="t>1")
        e1a1 = EventAssignment(variable="A", expression="3")
        e1a2 = EventAssignment(variable="B", expression="4")
        test_event=Event(name="e1", trigger=e1t, assignments=[e1a1, e1a2])
        model.add_event(test_event)
        passed_test = str(model)

    def test_run_nonsolver(self):
        model = Model()
        rate = Parameter(name = 'rate', expression = 0.5)
        model.add_parameter(rate)
        species1 = Species(name = 'A', initial_value = 0)
        species2 = Species(name = 'B', initial_value = 0)
        model.add_species(species1)
        model.add_species(species2)
        reaction = Reaction(name = 'reaction1', reactants={species1: 1}, products={species2: 1}, rate=rate)
        with self.assertRaises(SimulationError):
            results = model.run(number_of_trajectories = 1, solver = 'non_solver', seed = 1)

    def test_model_init_population_false_and_volume_warninag(self):
        with self.assertRaises(Warning):
            model = Model(population = False, volume = 0.9)

    def test_model_init_custom_tspan(self):
        model = Model(tspan = np.linspace(0, 20, 401))

    def test_parameter_init_unspecified_expression(self):
        with self.assertRaises(ParameterError):
            parameter = Parameter(name = 'parameter')

    def test_ode_propensity(self):
        model = Model()
        rate = Parameter(name='rate', expression=0.5)
        model.add_parameter(rate)
        species1 = Species('A', initial_value=10)
        species2 = Species('B', initial_value=10)
        model.add_species([species1, species2])
        r1 = Reaction(name='r1', reactants={'A':1}, products={}, rate=rate)
        r2 = Reaction(name='r2', reactants={'A':2}, products={'B':1}, rate=rate)
        r3 = Reaction(name='r3', reactants={'A':1, 'B':1}, products={}, rate=rate)
        r4 = Reaction(name='r4', reactants={'A':1}, products={}, propensity_function='t')
        model.add_reaction([r1, r2, r3, r4])
        self.assertEqual(model.listOfReactions['r1'].ode_propensity_function, 'rate*A')
        self.assertEqual(model.listOfReactions['r2'].ode_propensity_function, 'rate*A*A')
        self.assertEqual(model.listOfReactions['r3'].ode_propensity_function, 'rate*A*B')
        self.assertEqual(model.listOfReactions['r4'].ode_propensity_function, 't')


    def test_species_setter(self):
        sp1 = Species('A',initial_value=10)
        sp1.set_initial_value(5)
        self.assertEqual(sp1.initial_value,5)
        with self.assertRaises(SpeciesError):
            sp1.set_initial_value(-1)
        sp2 = Species('B',initial_value=5,mode='discrete')
        with self.assertRaises(SpeciesError):
            sp2.set_initial_value(.5)

    def test_robust_model(self):
        try:
            model = RobustModel()
            model.run()
        
        except ModelError as e:
            self.fail(f"Failed to instantiate the RobustModel: {e}")
        
        except SolverError as e:
            self.fail(f"Failed to run the RobustModel: {e}")

        except Exception as e:
            self.fail(f"An unknown exception occured while testing the RobustModel: {e}")

    def test_stochss_export(self):
        model = RobustModel()
        tempdir = tempfile.mkdtemp()
        stochss_model_path = os.path.join(tempdir, "robust_model.mdl")
        try:
            export_StochSS(model, filename=stochss_model_path)
            self.assertTrue(os.path.exists(stochss_model_path))
        finally:
            os.unlink(stochss_model_path)
            os.rmdir(tempdir)

    def test_run_example__with_increment_only(self):
        model = ExampleNoTspan()
        results = model.run(t=20, increment=0.2)

    def test_run_example__with_tspan_only(self):
        model = Example()
        results = model.run()

    def test_run_example__with_tspan_and_increment(self):
        with self.assertRaises(SimulationError):
            model = Example()
            results = model.run(t=20, increment=0.2)

if __name__ == '__main__':
    unittest.main()



"""
GillesPy2 is a modeling toolkit for biochemical simulation.
Copyright (C) 2019-2021 GillesPy2 developers.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import ast
try:
    import libsbml
except ImportError:
    raise ImportError('libsbml is required to convert GillesPy models to SBML files.')


def __get_math(math):
    replacements = {
            "log": "ln",
            "**": "^",
            "and": "&&",
            "or": "||"
            }
    for old, new in replacements.items():
        math_str = math.replace(old, new)
    return libsbml.parseL3Formula(math_str)

def __add_species(species_list, model):
    for name, species in species_list.items():
        spec = model.createSpecies()
        spec.initDefaults()
        spec.setCompartment("vol")
        spec.setId(name)
        spec.setInitialAmount(species.initial_value)

def __add_parameters(parameter_list, model):
    for name, parameter in parameter_list.items():
        param = model.createParameter()
        param.initDefaults()
        param.setId(name)
        try:
            param.setValue(ast.literal_eval(parameter.expression))
        except ValueError:
            param.setValue(parameter.expression)

def __add_reactions(reaction_list, model):
    for name, reaction in reaction_list.items():
        reac = model.createReaction()
        reac.initDefaults()
        reac.setId(name)

        __add_reactants(reaction.reactants, reac)
        __add_products(reaction.products, reac)

        propensity = __get_math(reaction.propensity_function)
        kin_law = reac.createKineticLaw()
        kin_law.setMath(propensity)

def __add_reactants(reactant_list, reaction):
    for spec, ratio in reactant_list.items():
        react = reaction.createReactant()
        react.setConstant(True)
        react.setSpecies(spec.name)
        react.setStoichiometry(ratio)

def __add_products(product_list, reaction):
    for spec, ratio in product_list.items():
        prod = reaction.createProduct()
        prod.setConstant(True)
        prod.setSpecies(spec.name)
        prod.setStoichiometry(ratio)

def __add_events(event_list, model):
    for name, event in event_list.items():
        evt = model.createEvent()
        evt.setId(name)
        evt.setUseValuesFromTriggerTime(event.use_values_from_trigger_time)
        
        if event.delay is not None:
            delay = __get_math(event.delay)
            dly = evt.createDelay()
            dly.setMath(delay)

        priority = __get_math(event.priority)
        prior = evt.createPriority()
        prior.setMath(priority)

        trigger = event.trigger
        expression = __get_math(trigger.expression)
        trig = evt.createTrigger()
        trig.setInitialValue(trigger.value)
        trig.setPersistent(trigger.persistent)
        trig.setMath(expression)

        __add_event_assignments(event.assignments, evt)

def __add_event_assignments(assignment_list, event):
    for assignment in assignment_list:
        assign = event.createEventAssignment()
        assign.setVariable(assignment.variable.name)

        expression = __get_math(assignment.expression)
        assign.setMath(expression)

def __add_rate_rules(rate_rule_list, model):
    for name, rule in rate_rule_list.items():
        r_rule = model.createRateRule()
        r_rule.setId(name)
        r_rule.setVariable(rule.variable)
        formula = __get_math(rule.formula)
        r_rule.setMath(formula)

def __add_assignment_rules(assignment_rule_list, model):
    for name, rule in assignment_rule_list.items():
        a_rule = model.createAssignmentRule()
        a_rule.setId(name)
        a_rule.setVariable(rule.variable)
        formula = __get_math(rule.formula)
        a_rule.setMath(formula)

def __add_function_definitions(function_definition_list, model):
    for name, function_def in function_definition_list.items():
        func_def = model.createFunctionDefinition()
        func_def.setId(name)
        func_str = f"lambda({function_def.get_arg_string()}, {function_def.function_string})"
        function = __get_math(func_str)
        func_def.setMath(function)

def __write_to_file(document, path):
    writer = libsbml.SBMLWriter()

    with open(path, "w") as sbml_file:
        sbml_file.write(writer.writeSBMLToString(document))

def export(model, path=None):
    """
    GillesPy model to SBML converter

    :param model: GillesPy model to be converted to SBML
    :type model: gillespy.Model

    :param path: Path to the SBML file for conversion
    :type path: str
    """

    if path is None:
        path = f"{model.name}.xml"

    document = libsbml.SBMLDocument(3, 2)

    sbml_model = document.createModel()
    sbml_model.setName(model.name)

    compartment = sbml_model.createCompartment()
    compartment.setId('vol')
    compartment.setConstant(True)
    compartment.setSize(model.volume)
    compartment.setSpatialDimensions(3)

    __add_species(model.listOfSpecies, sbml_model)
    __add_parameters(model.listOfParameters, sbml_model)
    __add_reactions(model.listOfReactions, sbml_model)
    __add_events(model.listOfEvents, sbml_model)
    __add_rate_rules(model.listOfRateRules, sbml_model)
    __add_assignment_rules(model.listOfAssignmentRules, sbml_model)
    __add_function_definitions(model.listOfFunctionDefinitions, sbml_model)

    __write_to_file(document, path)

    return path



"""
GillesPy2 is a modeling toolkit for biochemical simulation.
Copyright (C) 2019-2021 GillesPy2 developers.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import unittest
from gillespy2.core.model import Model


class EmptyModel(Model):
    def __init__(self, parameter_values=None):
        Model.__init__(self)


class TestEmptyModel(unittest.TestCase):
    def setUp(self):
        self.model = EmptyModel()

    def test_model_creation(self):
        name = self.model.name
        self.assertEqual(name, "", msg="Unexpected value: {}".format(name))


if __name__ == '__main__':
    unittest.main()



"""
GillesPy2 is a modeling toolkit for biochemical simulation.
Copyright (C) 2019-2021 GillesPy2 developers.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

import unittest
import tempfile
from gillespy2.core.gillespyError import DirectoryError, SimulationError
from example_models import Example, ExampleNoTspan
from gillespy2.solvers.cpp.tau_leaping_c_solver import TauLeapingCSolver

class TestTauLeapingCSolver(unittest.TestCase):
    def test_create(self):
        model = Example()
        solver = TauLeapingCSolver(model)

    def test_file_with_directory_name_exists(self):
        with self.assertRaises(DirectoryError):
            temp = tempfile.NamedTemporaryFile()
            model = Example()
            solver = TauLeapingCSolver(model, temp.name)

    def test_run_example_precompiled(self):
        model = Example()
        solver = TauLeapingCSolver(model)
        results = model.run(solver=solver)

    def test_run_example(self):
        model = Example()
        results = model.run(solver=TauLeapingCSolver)

    def test_run_example__with_increment_only(self):
        model = ExampleNoTspan()
        results = TauLeapingCSolver.run(model=model, increment=0.2)

    def test_run_example__with_tspan_only(self):
        model = Example()
        results = TauLeapingCSolver.run(model=model)

    def test_run_example__with_tspan_and_increment(self):
        with self.assertRaises(SimulationError):
            model = Example()
            results = TauLeapingCSolver.run(model=model, increment=0.2)


if __name__ == '__main__':
    unittest.main()



"""
GillesPy2 is a modeling toolkit for biochemical simulation.
Copyright (C) 2019-2021 GillesPy2 developers.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from gillespy2.core.gillespyError import *
from gillespy2.core.jsonify import Jsonify


class EventAssignment(Jsonify):
    """
    An EventAssignment describes a change to be performed to the current model
    simulation.  This is assignment can either be fired at the time its
    associated trigger changes from false to true, or after a specified delay,
    depending on how the Event to which it is assigned is configured.

    :param variable: Target model component to be modified by the EventAssignment
        expression. Valid target variables include gillespy2 Species,
        Parameters, and Compartments.
    :type variable: gillespy2.Species, gillespy2.Parameter

    :param expression: String to be evaluated when the event is fired.  This expression must
        be evaluable within the model namespace, and the results of it's
        evaluation will be assigned to the EventAssignment variable.
    :type expression: str

    """

    def __init__(self, variable=None, expression=None):

        self.variable = variable
        self.expression = expression

        if expression is not None:
            self.expression = str(expression)

        from gillespy2.core.parameter import Parameter
        from gillespy2.core.species import Species
        #TODO: ADD Compartment to valid variable types once implemented
        valid_variable_types = [Species, Parameter, str]

        if not type(variable) in valid_variable_types:
            print(variable)
            print(type(variable))
            raise EventError(
                'GillesPy2 Event Assignment variable must be a valid gillespy2 species')
        if not isinstance(self.expression, str):
            raise EventError(
                             'GillesPy2 Event Assignment expression requires a '
                             'valid string expression')

    def __str__(self):
        return self.variable.name + ': ' + self.expression

class EventTrigger(Jsonify):
    """
    Trigger detects changes in model/environment conditions in order to fire an
    event.  A Trigger contains an expression, a mathematical function which can
    be evaluated to a boolean value within a model's namespace.  Upon
    transitioning from 'false' to 'true', this trigger will cause the immediate
    execution of an event's list of assignments if no delay is present, otherwise,
    the delay evaluation will be initialized.

    :param expression: String for a function calculating EventTrigger values. Should be evaluable
        in namespace of Model.
    :type expression: str

    :param value: Value of EventTrigger at simulation start, with time t=0
    :type value: bool

    :param persistent: Determines if trigger condition is persistent or not
    :type persistent: bool
    """
    def __init__(self, expression=None, initial_value = False, persistent = False):

        if isinstance(expression, str):
            self.expression = expression
        else:
            raise EventError('EventTrigger expression must be a string')

        if isinstance(initial_value, bool):
            self.value = initial_value
        else:
            raise EventError('EventTrigger initial_value must be bool')

        if isinstance(persistent, bool):
            self.persistent = persistent
        else:
            raise EventError('EventTrigger.persistent must be bool')
    def __str__(self):
        return self.expression

    def sanitized_expression(self, species_mappings, parameter_mappings):
        names = sorted(list(species_mappings.keys()) + list(parameter_mappings.keys()), key=lambda x: len(x),
                       reverse=True)
        replacements = [parameter_mappings[name] if name in parameter_mappings else species_mappings[name]
                        for name in names]
        sanitized_expression = self.expression
        for id, name in enumerate(names):
            sanitized_expression = sanitized_expression.replace(name, "{"+str(id)+"}")
        return sanitized_expression.format(*replacements)


class Event(Jsonify):
    """
    An Event can be given as an assignment_expression (function) or directly
    as a value (scalar). If given an assignment_expression, it should be
    understood as evaluable in the namespace of a parent Model.

    :param name: The name by which this Event is called or referenced in reactions.
    :type name: str

    :param assignments: List of EventAssignments to be executed at trigger or delay
    :type assignments: str

    :param trigger: contains math expression which can be evaluated to
        a boolean result.  Upon the transition from 'False' to 'True',
        event assignments may be executed immediately, or after a
        designated delay.
    :type trigger: EventTrigger

    :param delay: contains math expression evaluable within model namespace.
        This expression designates a delay between the trigger of
        an event and the execution of its assignments.
    :type delay: str

    :param priority: Contains math expression evaluable within model namespace.
    :type priority: str

    :type use_values_from_trigger_time: bool
    """

    def __eq__(self, other):
        return str(self) == str(other)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __gt__(self, other):
        return not self.__le__(other)

    def __ge__(self, other):
        return not self.__lt__(other)

    def __lt__(self, other):
        return str(self) < str(other)

    def __le__(self, other):
        return str(self) <= str(other)

    def __hash__(self):
        if hasattr(self, '_hash'):
            return self._hash
        if hasattr(self, 'id'):
            self._hash = hash(self.id)
        elif hasattr(self, 'name'):
            self._hash = hash(self.name)
        else:
            self._hash = hash(self)
        return self._hash

    def __init__(self, name="", delay=None, assignments=[], priority="0", trigger=None,
                 use_values_from_trigger_time=False):

        # Events can contain any number of assignments
        self.assignments = []

        # Name
        if isinstance(name, str):
            self.name = name
        else:
            raise EventError(
             'name must be a valid string')
        
        # Trigger
        if hasattr(trigger, 'expression'):
            self.trigger = trigger
        else:
            raise EventError(
             'trigger must be set to a valid EventTrigger')
        
        # Delay
        if delay is None or isinstance(delay, str):
            self.delay = delay
        else:
            raise EventError(
             'delay must be a valid string or None')

        # Priority
        self.priority = priority

        # Assignments
        if isinstance(assignments, list):
            for assign in assignments:
                if hasattr(assign, 'variable'):
                    self.assignments.append(assign)
                else:
                    raise EventError('assignment list contains an item is not an EventAssignment.')
        elif hasattr(assignments, 'variable'):
            self.assignments.append(assignments)
        else:
            raise EventError(
                'assignments must contain only EventAssignments '
                'or a list of EventAssignments')
        # Use Values from Trigger Time
        if isinstance(use_values_from_trigger_time, bool):
            self.use_values_from_trigger_time = use_values_from_trigger_time
        else:
            raise EventError(
                'use_values_from_trigger_time requires bool')

    def __str__(self):
        print_string = self.name
        print_string += '\n\tTrigger: ' + str(self.trigger)
        if len(self.assignments):
            print_string += '\n\tAssignments:'
            for a in self.assignments:
                if isinstance(a.variable, str):
                    print_string += '\n\t\t' + a.variable + ': ' + a.expression
                else:
                    print_string += '\n\t\t' + a.variable.name + ': ' + a.expression
        return print_string

    def add_assignment(self, assignment):
        """
        Adds an EventAssignment or a list of EventAssignment.

        :param assignment: The event or list of events to be added to this event.
        :type assignment: EventAssignment or list[EventAssignment]
        """

        if hasattr(assignment, 'variable'):
            self.assignments.append(assignment)
        elif isinstance(assignment, list):
            for assign in assignment:
                if hasattr(assign, 'variable'):
                    self.assignments.append(assign)
                else:
                    raise EventError('add_assignment failed to add EventAssignment. Assignment to be added must be of '
                                     'type EventAssignment or list of EventAssignment objects.')
        else:
            raise ModelError("Unexpected parameter for add_assignment. Parameter must be EventAssignment or list of "
                             "EventAssignments")
        return assignment


"""
GillesPy2 is a modeling toolkit for biochemical simulation.
Copyright (C) 2019-2021 GillesPy2 developers.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from .tau_leaping_solver import TauLeapingSolver


class BasicTauLeapingSolver(TauLeapingSolver):
    """
    This solver has been deprecated. Future versions of GillesPy2 will not allow this import.
    Please import :mod:`gillespy2.solvers.numpy.TauLeapingSolver`.

    .. deprecated:: 1.5.0
    """

    from gillespy2.core import log
    log.warning("The name 'BasicTauLeapingSolver' has been deprecated, future versions of GillesPy2 will not allow"
                " this import. Please import 'TauLeapingSolver'")



"""
GillesPy2 is a modeling toolkit for biochemical simulation.
Copyright (C) 2019-2021 GillesPy2 developers.

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from threading import Thread, Event
from gillespy2.core.results import Results
from gillespy2.core import GillesPySolver, log, gillespyError
from gillespy2.solvers.utilities import solverutils as nputils
import random
import math
import numpy as np
np.set_printoptions(suppress=True)


class NumPySSASolver(GillesPySolver):
    name = "NumPySSASolver"
    rc = 0
    stop_event = None
    result = None
    pause_event = None

    def __init__(self, model=None):
        name = 'NumPySSASolver'
        rc = 0
        stop_event = None
        result = None
        pause_event = None
        self.model = model

    def get_solver_settings(self):
        """
        :returns: Tuple of strings, denoting all keyword argument for this solvers run() method.
        """
        return ('model', 't', 'number_of_trajectories', 'increment', 'seed', 'debug', 'timeout')

    @classmethod
    def run(self, model=None, t=20, number_of_trajectories=1, increment=None, seed=None, debug=False, show_labels=True,
            live_output=None, live_output_options={}, timeout=None, resume=None, **kwargs):

        """
        Run the SSA algorithm using a NumPy for storing the data in arrays and generating the timeline.

        :param model: The model on which the solver will operate.
        :param t: The end time of the solver.
        :param number_of_trajectories: The number of times to sample the chemical master equation. Each
            trajectory will be returned at the end of the simulation.
        :param increment: The time step of the solution.
        :param seed: The random seed for the simulation. Defaults to None.
        :param debug: Set to True to provide additional debug information about the
            simulation.
        :param resume: Result of a previously run simulation, to be resumed
        :param live_output: str The type of output to be displayed by solver. Can be "progress", "text", or "graph".
        :param live_output_options: dictionary contains options for live_output. By default {"interval":1}.
            "interval" specifies seconds between displaying.
            "clear_output" specifies if display should be refreshed with each display

        :returns: a list of each trajectory simulated.
        """

        if isinstance(self, type):
            self = NumPySSASolver(model=model)

        increment = self.get_increment(model=model, increment=increment)

        self.stop_event = Event()
        self.pause_event = Event()

        if timeout is not None and timeout <= 0:
            timeout = None
        if len(kwargs) > 0:
            for key in kwargs:
                log.warning('Unsupported keyword argument to {0} solver: {1}'.format(self.name, key))

        # create numpy array for timeline
        if resume is not None:
            # start where we last left off if resuming a simulation
            lastT = resume['time'][-1]
            step = lastT - resume['time'][-2]
            timeline = np.arange(lastT, t + step, step)
        else:
            timeline = np.linspace(0, t, int(round(t / increment + 1)))

        species = list(model._listOfSpecies.keys())

        trajectory_base, tmpSpecies = nputils.numpy_trajectory_base_initialization(model, number_of_trajectories,
                                                                                   timeline, species, resume=resume)

        # curr_time and curr_state are list of len 1 so that __run receives reference
        if resume is not None:
            total_time = [resume['time'][-1]]
        else:
            total_time = [0]

        curr_state = [None]
        live_grapher = [None]

        sim_thread = Thread(target=self.___run, args=(model, curr_state, total_time, timeline, trajectory_base,
                                                      live_grapher,), kwargs={'t': t, 'number_of_trajectories':
                                                                              number_of_trajectories,
                                                                              'increment': increment,
                                                                              'seed': seed, 'debug': debug,
                                                                              'show_labels': show_labels,
                                                                              'timeout': timeout,
                                                                              'resume': resume, })
        try:
            time = 0
            sim_thread.start()
            if live_output is not None:
                import gillespy2.core.liveGraphing
                live_output_options['type'] = live_output
                gillespy2.core.liveGraphing.valid_graph_params(live_output_options)
                if resume is not None:
                    resumeTest = True  # If resuming, relay this information to live_grapher
                else:
                    resumeTest = False
                live_grapher[0] = gillespy2.core.liveGraphing.LiveDisplayer(model, timeline, number_of_trajectories,
                                                                             live_output_options,resume = resumeTest)
                display_timer = gillespy2.core.liveGraphing.RepeatTimer(live_output_options['interval'],
                                                                        live_grapher[0].display, args=(curr_state,
                                                                                                       total_time,
                                                                                                       trajectory_base,
                                                                                                       live_output
                                                                                                       )
                                                                        )
                display_timer.start()

            if timeout is not None:
                while sim_thread.is_alive():
                    sim_thread.join(.1)
                    time += .1
                    if time >= timeout:
                        break
            else:
                while sim_thread.is_alive():
                    sim_thread.join(.1)

            if live_grapher[0] is not None:
                display_timer.cancel()
            self.stop_event.set()
            while self.result is None:
                pass
        except KeyboardInterrupt:
            if live_output:
                display_timer.pause = True
                display_timer.cancel()
            self.pause_event.set()
            while self.result is None:
                pass
        if hasattr(self, 'has_raised_exception'):
            raise self.has_raised_exception

        return Results.build_from_solver_results(self, live_output_options)

    def ___run(self, model, curr_state, total_time, timeline, trajectory_base, live_grapher, t=20,
               number_of_trajectories=1, increment=0.05, seed=None, debug=False, show_labels=True, resume=None,
               timeout=None):

        try:
            self.__run(model, curr_state, total_time, timeline, trajectory_base, live_grapher, t, number_of_trajectories,
                       increment, seed, debug, show_labels, resume, timeout)
        except Exception as e:
            self.has_raised_exception = e
            self.result = []
            return [], -1

    def __run(self, model, curr_state, total_time, timeline, trajectory_base, live_grapher, t=20,
              number_of_trajectories=1, increment=0.05, seed=None, debug=False, show_labels=True,
              resume=None,  timeout=None):

        # for use with resume, determines how much excess data to cut off due to
        # how species and time are initialized to 0
        timeStopped = 0

        if resume is not None:
            if resume[0].model != model:
                raise gillespyError.ModelError('When resuming, one must not alter the model being resumed.')
            if t < resume['time'][-1]:
                raise gillespyError.ExecutionError(
                    "'t' must be greater than previous simulations end time, or set in the run() method as the "
                    "simulations next end time")

        random.seed(seed)

        species_mappings, species, parameter_mappings, number_species = nputils.numpy_initialization(model)

        # create dictionary of all constant parameters for propensity evaluation
        parameters = {'V': model.volume}
        for paramName, param in model.listOfParameters.items():
            parameters[parameter_mappings[paramName]] = param.value

        # create mapping of reaction dictionary to array indices
        reactions = list(model.listOfReactions.keys())

        # create mapping of reactions, and which reactions depend on their reactants/products
        dependent_rxns = nputils.dependency_grapher(model, reactions)
        number_reactions = len(reactions)
        propensity_functions = {}

        # create an array mapping reactions to species modified
        species_changes = np.zeros((number_reactions, number_species))

        # pre-evaluate propensity equations from strings:
        for i, reaction in enumerate(reactions):
            # replace all references to species with array indices
            for j, spec in enumerate(species):
                species_changes[i][j] = model.listOfReactions[reaction].products.get(model.listOfSpecies[spec], 0) \
                                        - model.listOfReactions[reaction].reactants.get(model.listOfSpecies[spec], 0)
                if debug:
                    print('species_changes: {0},i={1}, j={2}... {3}'.format(species, i, j, species_changes[i][j]))
            propensity_functions[reaction] = [eval('lambda S:' + model.listOfReactions[reaction].
                                                   sanitized_propensity_function(species_mappings, parameter_mappings),
                                                   parameters), i]
        if debug:
            print('propensity_functions', propensity_functions)

        # begin simulating each trajectory
        simulation_data = []
        for trajectory_num in range(number_of_trajectories):
            if self.stop_event.is_set():
                self.rc = 33
                break
            elif self.pause_event.is_set():
                timeStopped = timeline[entry_count]
                break

            # For multi trajectories, live_grapher needs to be informed of trajectory increment
            if live_grapher[0] is not None:
                live_grapher[0].increment_trajectory(trajectory_num)

            # copy initial state data
            trajectory = trajectory_base[trajectory_num]
            entry_count = 1
            curr_state[0] = {}
            # curr_time and curr_state are list of len 1 so that __run receives reference
            if resume is not None:
                curr_time = [resume['time'][-1]]
            else:
                curr_time = [0]

            for spec in model.listOfSpecies:
                if resume is not None:
                    curr_state[0][spec] = resume[spec][-1]
                else:
                    curr_state[0][spec] = model.listOfSpecies[spec].initial_value

            propensity_sums = np.zeros(number_reactions)
            # calculate initial propensity sums
            while entry_count < timeline.size:
                if self.stop_event.is_set():
                    self.rc = 33
                    break
                elif self.pause_event.is_set():
                    timeStopped = timeline[entry_count]
                    break
                # determine next reaction

                species_states = list(curr_state[0].values())

                for i in range(number_reactions):
                    propensity_sums[i] = propensity_functions[reactions[i]][0](species_states)

                    if debug:
                        print('propensity: ', propensity_sums[i])

                propensity_sum = np.sum(propensity_sums)
                if debug:
                    print('propensity_sum: ', propensity_sum)
                # if no more reactions, quit
                if propensity_sum <= 0:
                    trajectory[entry_count:, 1:] = list(species_states)
                    break

                cumulative_sum = random.uniform(0, propensity_sum)
                rand = random.random()
                curr_time[0] += -math.log(rand) / propensity_sum
                total_time[0] += -math.log(rand) / propensity_sum
                if debug:
                    print('cumulative sum: ', cumulative_sum)
                    print('entry count: ', entry_count)
                    print('timeline.size: ', timeline.size)
                    print('curr_time: ', curr_time[0])
                # determine time passed in this reaction

                while entry_count < timeline.size and timeline[entry_count] <= curr_time[0]:
                    if self.stop_event.is_set():
                        self.rc = 33
                        break
                    elif self.pause_event.is_set():
                        timeStopped = timeline[entry_count]
                        break
                    trajectory[entry_count, 1:] = species_states
                    entry_count += 1

                for potential_reaction in range(number_reactions):
                    cumulative_sum -= propensity_sums[potential_reaction]
                    if debug:
                        print('if <=0, fire: ', cumulative_sum)
                    if cumulative_sum <= 0:

                        for i,spec in enumerate(model.listOfSpecies):
                            curr_state[0][spec] += species_changes[potential_reaction][i]

                        reacName = reactions[potential_reaction]
                        if debug:
                            print('current state: ', curr_state[0])
                            print('species_changes: ', species_changes)
                            print('updating: ', potential_reaction)

                        species_states = list(curr_state[0].values())
                        for i in dependent_rxns[reacName]['dependencies']:
                            propensity_sums[propensity_functions[i][1]] = propensity_functions[i][0](species_states)

                            if debug:
                                print('new propensity sum: ', propensity_sums[i])
                        break
            data = {
                'time': timeline
            }
            for i in range(number_species):
                data[species[i]] = trajectory[:, i+1]
            simulation_data.append(data)

        # If simulation has been paused, or tstopped !=0
        if timeStopped != 0 or resume is not None:
            simulation_data = nputils.numpy_resume(timeStopped, simulation_data, resume=resume)

        self.result = simulation_data
        return self.result, self.rc
